"""Contain all function representation of jmc terminal command"""
from datetime import datetime
from json import dump
import os
from pathlib import Path
import sys
import threading
from time import perf_counter
from traceback import format_exc

from .terminal.utils import RestartException, error_report, get_input, handle_exception, press_enter
from .terminal import pprint, Colors, GlobalData, add_command
from .compile import compile_jmc, Logger, EXCEPTIONS, get_debug_log, get_info_log

global_data = GlobalData()
logger = Logger(__name__)

NEW_LINE = "\n"


@add_command("help [<command>]", rename="help")
def help_(command: str = "") -> None:
    """Output this message or get help for a command"""
    if not command:
        avaliable_commands = "\n".join(f" - {usage}: {func.__doc__}" for _,
                                       (func, usage) in global_data.commands.items())
        pprint(f"""Avaliable commands:

{avaliable_commands}
""", color=Colors.YELLOW)
        return

    if command not in global_data.commands:
        raise TypeError(f"Unrecognized command '{command}'")
    func, usage = global_data.commands[command]
    pprint(f"{usage}: {func.__doc__}", Colors.INFO)


@add_command("exit", rename="exit")
def exit_() -> None:
    """Exit JMC compiler"""
    sys.exit(0)


@add_command("compile [debug]", "compile")
def compile_(debug: str = "") -> None:
    """Compile main JMC file"""
    if debug:
        if debug != 'debug':
            raise TypeError(f"Unrecognized argument '{debug}'")
        pprint("DEBUG MODE", Colors.INFO)
    debug_compile = bool(debug)

    pprint("Compiling...", Colors.INFO)
    try:
        start_time = perf_counter()
        compile_jmc(global_data.config, debug=True)
        stop_time = perf_counter()
        pprint(
            f"Compiled successfully in {stop_time-start_time} seconds", Colors.INFO)
    except EXCEPTIONS as error:
        logger.debug(format_exc())
        error_report(error)
    except Exception as error:
        logger.exception("Non-JMC Error occur")
        handle_exception(error, global_data.EVENT, is_ok=False)

    if debug_compile:
        __log_debug()
        __log_info()


def __log_debug() -> None:
    logger.info("Requesting debug log")
    global_data.LOG_PATH.mkdir(exist_ok=True)
    debug_log = get_debug_log()
    with (global_data.LOG_PATH / datetime.now().strftime("JMC_DEBUG - %y-%m-%d %H.%M.%S.log")).open('w+') as file:
        file.write(debug_log)
    with (global_data.LOG_PATH / "latest.log").open('w+') as file:
        file.write(debug_log)


def __log_info() -> None:
    logger.info("Requesting info log")
    global_data.LOG_PATH.mkdir(exist_ok=True)
    info_log = get_info_log()
    with (global_data.LOG_PATH / datetime.now().strftime("JMC_INFO - %y-%m-%d %H.%M.%S.log")).open('w+') as file:
        file.write(info_log)
    with (global_data.LOG_PATH / "latest.log").open('w+') as file:
        file.write(info_log)


def __log_clear() -> None:
    logger.info("Clearing logs")
    if not global_data.LOG_PATH.is_dir():
        logger.debug("Log folder not found")
        return

    for path in global_data.LOG_PATH.iterdir():
        if not path.is_file():
            continue
        if path.suffix != '.log':
            continue
        if path.name == 'latest.log':
            continue
        path.unlink()


@add_command("log debug|info|clear")
def log(mode: str) -> None:
    """
debug|info: Create log file in output directory
clear: Delete every log file inside log folder except latest"""
    if mode not in {"debug", "info", "clear"}:
        raise TypeError(f"Unrecognized mode '{mode}'")

    if mode == "debug":
        __log_debug()
    elif mode == "info":
        __log_info()
    elif mode == "clear":
        __log_clear()


@add_command("version")
def version() -> None:
    """Get JMC's version info"""
    pprint(global_data.VERSION, Colors.INFO)


def __config_reset() -> None:
    (global_data.cwd / global_data.CONFIG_FILE_NAME).unlink(missing_ok=True)
    pprint("Resetting configurations", Colors.PURPLE)
    print('\n' * 5)
    raise RestartException()


def __config_edit() -> None:
    config_json = global_data.config.toJSON()
    pprint(f"""Edit configurations (Bypass error checking)
Type `cancel` to cancel
{NEW_LINE.join([f"- {key}" for key in config_json])}""", Colors.PURPLE)
    key = get_input("Configuration: ")
    if key not in config_json:
        if key.lower() == 'cancel':
            return
        pprint("Invalid Key", Colors.FAIL)
        __config_edit()
    else:
        pprint(f"Current {key}: {config_json[key]}", Colors.YELLOW)
        config_json[key] = get_input("New Value: ")
        with (global_data.cwd / global_data.CONFIG_FILE_NAME).open('w') as file:
            dump(config_json, file, indent=2)

        global_data.config.load_config()


@add_command("config edit|reset")
def config(mode: str) -> None:
    """
reset: Delete the configuration file and restart the compiler
edit: Override configuration file and bypass error checking"""
    if mode == 'reset':
        __config_reset()
    elif mode == 'edit':
        __config_edit()
    else:
        raise TypeError(f"Unrecognized mode '{mode}'")


def __background() -> None:
    while not global_data.EVENT.is_set():
        logger.debug("Auto compiling")
        compile_()
        global_data.EVENT.wait(global_data.interval)


@add_command("autocompile <interval (second)>")
def autocompile(interval: str) -> None:
    """Start automatically compiling with certain interval (Press Enter to stop)"""
    try:
        global_data.interval = int(interval)
    except ValueError as error:
        raise TypeError("Invalid integer for interval") from error
    if global_data.interval == 0:
        pprint("Interaval cannot be 0 seconds", Colors.FAIL)
        return

    thread = threading.Thread(
        target=__background,
        daemon=True
    )
    global_data.EVENT.clear()
    thread.start()

    press_enter("Press Enter to stop...\n")
    pprint("Stopping...", Colors.INFO)
    global_data.EVENT.set()
    thread.join()


@add_command("cd <path>", rename="cd")
def chdir(path: str) -> None:
    """Change current directory"""
    try:
        os.chdir(path)
    except ValueError:
        pprint("Invalid path", Colors.FAIL)
        return
    global_data.cwd = Path(os.getcwd())
    global_data.LOG_PATH = global_data.cwd / 'log'
    raise RestartException



from typing import Callable
from ._flow_control import (
    if_,
    else_,
    while_,
    do,
    switch,
    for_
)

from ..tokenizer import Token, Tokenizer
from ..datapack import DataPack

FLOW_CONTROL_COMMANDS: dict[str, Callable[
    [
        list[Token],
        DataPack,
        Tokenizer,
    ], str | None]] = {
    'while': while_,
    'do': do,
    'if': if_,
    'else': else_,
    'switch': switch,
    'for': for_,
}



"""Module handling datapack"""
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Iterable
from json import JSONEncoder, dumps


from .tokenizer import Token, TokenType, Tokenizer
from .datapack_data import Data
from .exception import JMCSyntaxWarning, JMCValueError
from .log import Logger

if TYPE_CHECKING:
    from .lexer import Lexer
    from .header import Header

logger = Logger(__name__)


NEW_LINE = '\n'


class FunctionEncoder(JSONEncoder):
    """
    Custom minecraft function encoder for json.dump
    """

    def default(self, o):
        if isinstance(o, Function):
            return o.commands
        return super().default(o)


class Function:
    """
    A class representation for a minecraft function (.mcfunction)

    :param commands: List of minecraft commands(string), defaults to empty list
    """
    __slots__ = ('commands', )
    commands: list[str]

    def __init__(self, commands: list[str] | None = None) -> None:
        if commands is None:
            self.commands = []
        else:
            self.commands = self.__split(commands)

    def add_empty_line(self) -> None:
        """
        Add empty line at the end of the function
        """
        self.commands.append('')

    def append(self, command: str) -> None:
        """
        Append 1 or more command in form of a single string

        :param command: 1 or more line of minecraft command string
        """
        self.commands.extend(self.__split([command]))

    def extend(self, commands: list[str]) -> None:
        """
        Append multiple minecraft commands

        :param commands: List of minecraft commands(strings), each string can have multiple lines
        """
        self.commands.extend(self.__split(commands))

    def insert_extend(self, commands: list[str], index: int) -> None:
        """
        Append multiple minecraft commands to a certine line of function

        :param commands:  List of minecraft commands(strings), each string can have multiple lines
        :param index: Line/Position of the function to insert to
        """
        self.commands[index:index] = self.__split(commands)

    def delete(self, index: int) -> None:
        """
        Delete a command

        :param index: Line/Position of the function to delete
        """
        del self.commands[index]

    @property
    def content(self) -> str:
        """
        Content of the function in 1 string

        :return: Commands of the function in form of a single string
        """
        return '\n'.join(self.commands)

    @property
    def length(self) -> int:
        """
        Amount of lines in the function (including empty lines)

        :return: Length of commands attribute
        """
        return len(self.commands)

    def __repr__(self) -> str:
        return f"Function({repr(self.commands)})"

    def __iter__(self) -> Iterable:
        return self.commands.__iter__()

    def __bool__(self):
        return bool(self.commands)

    def __split(self, strings: list[str]) -> list[str]:
        """
        Loop through every line in each string of command(s) and make a new list with every element having only 1 line of command

        :param strings: minecraft commands(strings),each string can have multiple lines
        :return: minecraft commands(strings),each string can have only a single line
        """
        return [
            line for string in strings for line in string.split('\n') if line]


class DataPack:
    """
    A class representation for entire minecraft datapack

    :param namespace: Datapack's namespace
    :param lexer: Lexer object
    """
    __slot__ = ('data', 'ints',
                'functions', 'load_function', 'jsons',
                'private_functions', 'private_function_count',
                '__scoreboards', 'loads', 'ticks', 'namespace',
                'used_command', 'lexer')
    private_name = '__private__'
    load_name = '__load__'
    tick_name = '__tick__'
    var_name = '__variable__'
    int_name = '__int__'
    VARIABLE_SIGN = '$'
    """Data read from header file(s)"""

    def __init__(self, namespace: str, lexer: "Lexer") -> None:
        logger.debug("Initializing Datapack")
        self.ints: set[int] = set()
        """Set of integers going to be used in scoreboard"""
        self.functions: dict[str, Function] = {}
        """Dictionary of function name and a Function object"""
        self.load_function: list[list[Token]] = []
        """List of commands(list of tokens) in load function"""
        self.jsons: dict[str, dict[str, Any] | list[Any]] = defaultdict(dict)
        """Dictionary of json name and a dictionary(jsobject)"""
        self.private_functions: dict[str,
                                     dict[str, Function]] = defaultdict(dict)
        """Dictionary of function's group name and (Dictionary of function name and a Function object)"""
        self.private_function_count: dict[str, int] = defaultdict(int)
        """Current count of how many private functions there are in each group name"""
        self.__scoreboards: dict[str, str] = {
            self.var_name: 'dummy',
            self.int_name: 'dummy'
        }
        """Minecraft scoreboards that are going to be created"""

        self.loads: list[str] = []
        """Output list of commands for load"""
        self.ticks: list[str] = []
        """Output list of commands for tick"""
        self.namespace = namespace
        """Datapack's namespace"""

        self.used_command: set[str] = set()
        """Used JMC command that's for one time call only"""

        self.lexer = lexer
        """Lexer object"""

        self.data = Data()
        """Extra information that can be shared across all JMC function"""

    def add_objective(self, objective: str, criteria: str = 'dummy') -> None:
        """
        Add minecraft scoreboard objective

        :param objective: Name of scoreboard
        :param criteria: Criteria of scoreboard, defaults to 'dummy'
        :raises ValueError: If scoreboard already exists
        """
        if objective in self.__scoreboards and self.__scoreboards[objective] != criteria:
            raise ValueError(
                f"Conflict on adding scoreboard, '{objective}' objective with '{self.__scoreboards[objective]}' criteria already exist.\nGot same objective with '{criteria}' criteria.")
        self.__scoreboards[objective] = criteria

    def get_count(self, name: str) -> str:
        """
        Get count as a string from private function's group

        :param name: Name of the function group
        :return: Count as a string
        """
        count = self.private_function_count[name]
        self.private_function_count[name] += 1
        return str(count)

    def call_func(self, name: str, count: str) -> str:
        """
        Get command string for calling minecraft private function

        :param name: Name of the private function group
        :param count: Name of the function (usually as count)
        :return: String for calling minecraft function
        """
        return f"function {self.namespace}:{self.private_name}/{name}/{count}"

    def add_private_json(self, json_type: str, name: str,
                         json: dict[str, Any] | list[Any]) -> None:
        """
        Add new private json to datapack

        :param json_type: Minecraft json type, for example: tags/functions
        :param name: Name of the private json
        :param json: Dictionary object
        """
        self.jsons[f"{json_type}/{self.private_name}/{name}"] = json

    def add_json(self, json_type: str, name: str,
                 json: dict[str, Any] | list[Any]) -> None:
        """
        Add new private json to datapack

        :param json_type: Minecraft json type, for example: tags/functions
        :param name: Name of the json
        :param json: Dictionary object
        """
        self.jsons[f"{json_type}/{name}"] = json

    def add_arrow_function(self, name: str, token: Token,
                           tokenizer: Tokenizer) -> str:
        """
        Add private function for user (arrow function)

        :param name: Private function's group name
        :param token: paren_curly token
        :param tokenizer: token's tokenizer
        :raises JMCSyntaxWarning: If the string in curly bracket is empty
        :return: Minecraft function call string
        """
        if token.string == '{}':
            raise JMCSyntaxWarning("Empty function", token, tokenizer)

        commands = self.parse_function_token(token, tokenizer)
        if len(commands) == 1 and NEW_LINE not in commands[0]:
            return commands[0]

        count = self.get_count(name)
        self.private_functions[name][count] = Function(commands)
        return self.call_func(name, count)

    def add_custom_private_function(self, name: str, token: Token, tokenizer: Tokenizer, count: str,
                                    precommands: list[str] | None = None, postcommands: list[str] | None = None) -> str:
        """
        Wrap custom commands around user's commands

        :param name: Private function's group name
        :param token: paren_curly token
        :param tokenizer: token's tokenizer
        :param count: Name of the function (usually as count)
        :param precommands: Commands before user's commands
        :param postcommands: Commands after user's commands
        :return: Minecraft function call string
        """
        if precommands is None:
            precommands = []
        if postcommands is None:
            postcommands = []

        commands = [*precommands,
                    *self.parse_function_token(token, tokenizer),
                    *postcommands]
        self.private_functions[name][count] = Function(commands)
        return self.call_func(name, count)

    def add_raw_private_function(
            self, name: str, commands: list[str], count: str | None = None) -> str:
        """
        Add private function for JMC as is

        :param name: Name of the private function's group
        :param commands: List of commands(string)
        :param count: Name of the function (usually as count), defaults to current count + 1
        :return: Minecraft function call string
        """
        if count is None:
            count = self.get_count(name)
        self.private_functions[name][count] = Function(commands)
        return self.call_func(name, count)

    def parse_function_token(self, token: Token,
                             tokenizer: Tokenizer) -> list[str]:
        """
        "Parse a paren_curly token into a list of commands(string)

        :param token: paren_curly token
        :param tokenizer: token's tokenizer
        :return: List of minecraft commands(string)
        """
        return self.lexer.parse_func_content(
            token.string[1:-1], tokenizer.file_path, token.line, token.col, tokenizer.file_string)

    def add_tick_command(self, command: str) -> None:
        """
        Add command to self.ticks

        :param command: Minecraft command string
        """
        self.ticks.append(command)

    def add_load_command(self, command: str) -> None:
        """
        Add command to self.loads

        :param command: Minecraft command string
        """
        self.loads.append(command)

    def add_int(self, integer: int) -> None:
        """
        Add command to self.ints

        :param command: Minecraft command string
        """
        self.ints.add(integer)

    def build(self) -> None:
        """
        Finializing DataPack for building (NO file writing)
        """
        logger.debug("Finializing DataPack")
        self.loads[0:0] = [
            *[f"scoreboard objectives add {objective} {criteria}" for objective,
                criteria in self.__scoreboards.items()],
            *[f"scoreboard players set {n} {self.int_name} {n}" for n in self.ints],
        ]
        if self.loads:
            self.functions[self.load_name].insert_extend(self.loads, 0)
        if self.ticks:
            if self.tick_name in self.functions:
                self.functions[self.tick_name].insert_extend(self.ticks, 0)
            else:
                self.functions[self.tick_name] = Function(self.ticks)
        for name, functions in self.private_functions.items():
            for path, func in functions.items():
                self.functions[f"{self.private_name}/{name}/{path}"] = func

        self.private_functions = {}
        self.loads = []
        self.ticks = []

    def parse_func_map(self, token: Token,
                       tokenizer: Tokenizer) -> dict[int, tuple[str, bool]]:
        """
        Parse JMC function hashmap

        :param token: paren_curly token
        :param tokenizer: token's tokenizer
        :param datapack: Datapack object
        :return: Dictionary of integer key and (tuple of function string and whether it is an arrow function)
        """
        func_map: dict[int, tuple[str, bool]] = {}
        for key, value in tokenizer.parse_js_obj(token).items():
            try:
                num = int(key)
            except ValueError as error:
                raise JMCValueError(
                    f"Expected number as key (got {key})", token, tokenizer) from error

            if value.token_type == TokenType.KEYWORD:
                func_map[num] = value.string, False
            elif value.token_type == TokenType.FUNC:
                func_map[num] = '\n'.join(
                    self.parse_function_token(value, tokenizer)), True
            else:
                raise JMCValueError(
                    f"Expected function, got {value.token_type.value}", token, tokenizer)
        return func_map

    def parse_list(self, token: Token,
                   tokenizer: Tokenizer, list_of: TokenType) -> list[str]:
        """
        Parse paren_square token into list of strings

        :param token: paren_square token
        :param tokenizer: token's Tokenizer
        :param list_of: TokenType of elements in the list for verification
        :raises JMCValueError: Wrong TokenType
        :return: List of strings
        """
        token_list = tokenizer.parse_list(token)
        for token_ in token_list:
            if token_.token_type != list_of:
                raise JMCValueError(
                    f"Expected list/array of {list_of.value}, got {token_.token_type.value}", token, tokenizer)
        return [token_.string for token_ in token_list]

    def token_dict_to_raw_js_object(
            self, token_dict: dict[str, Token], tokenizer: Tokenizer) -> str:
        """
        Turns a dictionary of key and token to a string in form of Object

        :param token_dict: Dictionary of string and Token
        :return: String that looks like object
        """
        pairs = []
        for key, token in token_dict.items():
            if token.token_type == TokenType.STRING:
                pairs.append(f"{key}:{token.add_quotation()}")
            elif token.token_type in {TokenType.PAREN_CURLY, TokenType.PAREN_ROUND, TokenType.PAREN_SQUARE}:
                pairs.append(
                    f"{key}:{self.lexer.clean_up_paren_token(token, tokenizer, is_nbt=False)}")
            else:
                pairs.append(f"{key}:{token.string}")
        return '{' + ",".join(pairs) + '}'

    def __repr__(self) -> str:
        return f"""DataPack(
    PRIVATE_NAME = {self.private_name},
    LOAD_NAME = {self.load_name},
    TICK_NAME = {self.tick_name},
    VAR_NAME = {self.var_name},
    INT_NAME = {self.int_name}

    objectives = {dumps(self.__scoreboards, indent=2)}
    ints = {self.ints!r}
    functions = {dumps(self.functions, indent=2, cls=FunctionEncoder)}
    jsons = {dumps(self.jsons, indent=2)}
    private_functions = {dumps(self.private_functions, indent=2, cls=FunctionEncoder)}
    loads = {dumps(self.loads, indent=2)}
    tick = {dumps(self.ticks, indent=2)}

)"""

# {dumps({key:list(value) for key, value in self.functions.items()}, indent=2)}



"""Main module"""
import atexit
from .compile import Logger
from .terminal.utils import RestartException, handle_exception
from .terminal import GlobalData, Colors, start

VERSION = 'v1.2.8-alpha.1'
CONFIG_FILE_NAME = 'jmc_config.json'
GlobalData().init(VERSION, CONFIG_FILE_NAME)

from . import terminal_commands  # noqa

logger = Logger(__name__)


def main():
    """Main function"""
    atexit.register(lambda: print(Colors.EXIT.value))
    logger.info("Starting session")
    global_data = GlobalData()
    while True:
        try:
            start()
        except Exception as error:
            handle_exception(error, global_data.EVENT, is_ok=True)
        except RestartException:
            pass
        except KeyboardInterrupt:
            break


if __name__ == '__main__':
    main()



"""Module containing all jmc exceptions"""
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING, Callable
from .log import Logger

if TYPE_CHECKING:
    from .tokenizer import Token, Tokenizer

logger = Logger(__name__)
NEW_LINE = '\n'


def log(self: object, args: tuple):
    """
    Log the exception as warning

    :param self: Exception itself
    :param args: Exception's arguments (Starting with message)
    """
    logger.warning(f"{self.__class__.__name__}\n{args[0]}")


def error_msg(message: str, token: "Token|None", tokenizer: "Tokenizer", col_length: bool,
              display_col_length: bool, entire_line: bool, suggestion: str | None, overide_file_str: Callable[[str], str] = lambda string: string) -> str:
    """
    Generate error message

    :param message: Main message at the front of the error
    :param token: A token/tokenizer where the error happens
    :param tokenizer: token's tokenizer
    :param col_length: Whether to add token's length to column count
    :param display_col_length: Whether to display the code until the end of the token
    :param entire_line: Whether to display the entire line of code
    :param suggestion: A suggestion message at the end of the error message
    :return: Error message
    """

    if token is None:
        string = ""
        length = 1
        col = tokenizer.col
        line = tokenizer.line
    else:
        string = token.get_full_string()
        length = token.length
        col = token.col
        line = token.line

    display_line = line
    display_col = col

    if col_length:
        if '\n' in string:
            line += string.count('\n')
            col = length - string.rfind('\n')
        else:
            col += length

    if display_col_length:
        if '\n' in string:
            display_line += string.count('\n')
            display_col = length - string.rfind('\n')
        else:
            display_col += length
    else:
        display_col += 1
    if entire_line:
        msg = f"In {tokenizer.file_path}\n{message} at line {line}.\n{overide_file_str(tokenizer.file_string.split(NEW_LINE)[display_line-1])} <-"
    else:
        msg = f"In {tokenizer.file_path}\n{message} at line {line} col {col}.\n{overide_file_str(tokenizer.file_string.split(NEW_LINE)[display_line-1])[:display_col-1]} <-"
    if suggestion is not None:
        msg += '\n' + suggestion
    return msg


class HeaderFileNotFoundError(FileNotFoundError):
    """Header file not found"""

    def __init__(self, path: Path):
        msg = f"Header file not found: {path.as_posix()}"
        log(self, (msg, ))
        super().__init__(msg)


class HeaderDuplicatedMacro(ValueError):
    """Define same macro twice"""

    def __init__(self, message: str, file_name: str, line: int, line_str: str):
        msg = f"In {file_name}\n{message} at line {line}\n{line_str}"
        log(self, (msg, ))
        super().__init__(msg)


class HeaderSyntaxException(SyntaxError):
    """Invalid syntax for header"""

    def __init__(self, message: str, file_name: str, line: int, line_str: str):
        msg = f"In {file_name}\n{message} at line {line}\n{line_str}"
        log(self, (message, ))
        super().__init__(msg)


class JMCSyntaxException(SyntaxError):
    """Invalid syntax for JMC"""

    def __init__(self, message: str, token: "Token|None", tokenizer: "Tokenizer", *, col_length: bool = False,
                 display_col_length: bool = True, entire_line: bool = False, suggestion: str | None = None) -> None:
        self.message = message
        self.token = token
        self.tokenizer = tokenizer
        self.col_length = col_length
        self.display_col_length = display_col_length
        self.entire_line = entire_line
        self.suggestion = suggestion
        msg = error_msg(message, token, tokenizer, col_length,
                        display_col_length, entire_line, suggestion)
        log(self, (msg, ))
        super().__init__(msg)

    def reinit(self, overide_file_str: Callable[[str], str]):
        self.msg = error_msg(self.message, self.token, self.tokenizer, self.col_length,
                             self.display_col_length, self.entire_line, self.suggestion, overide_file_str)
        """
        Overide initiation
        """
        logger.warning("Overiding file string")
        log(self, (self.msg, ))


class JMCValueError(ValueError):
    """Invalid syntax for JMC"""

    def __init__(self, message: str, token: "Token|None", tokenizer: "Tokenizer", *, col_length: bool = False,
                 display_col_length: bool = True, entire_line: bool = False, suggestion: str | None = None) -> None:
        msg = error_msg(message, token, tokenizer, col_length,
                        display_col_length, entire_line, suggestion)
        log(self, (msg, ))
        super().__init__(msg)


class JMCSyntaxWarning(SyntaxWarning):
    """Warnings about dubious JMC syntax"""

    def __init__(self, message: str, token: "Token|None", tokenizer: "Tokenizer", *, col_length: bool = False,
                 display_col_length: bool = True, entire_line: bool = False, suggestion: str | None = None) -> None:
        msg = error_msg(message, token, tokenizer, col_length,
                        display_col_length, entire_line, suggestion)
        log(self, (msg, ))
        super().__init__(msg)


class JMCFileNotFoundError(FileNotFoundError):
    """JMC file not found"""

    def __init__(self, *args: object) -> None:
        log(self, args)
        super().__init__(*args)


class JMCBuildError(Exception):
    """Cannot build the datapack"""

    def __init__(self, *args: object) -> None:
        log(self, args)
        super().__init__(*args)


class JMCDecodeJSONError(ValueError):
    """Invalid syntax for JSON"""

    def __init__(self, error: JSONDecodeError, token: "Token",
                 tokenizer: "Tokenizer") -> None:
        line = token.line + error.lineno - 1
        col = token.col + error.colno - 1 \
            if token.line == line else error.colno

        msg = f"In {tokenizer.file_path}\n{error.msg} at line {line} col {col}.\n{tokenizer.file_string.split(NEW_LINE)[line-1][:col-1]} <-"

        log(self, (msg, ))
        super().__init__(msg)


class JMCMissingValueError(ValueError):
    """Missing required positional argument"""

    def __init__(self, missing_argument: str, token: "Token",
                 tokenizer: "Tokenizer", *, suggestion: str | None = None) -> None:
        msg = error_msg(f"Missing required positional argument: '{missing_argument}'", token, tokenizer, col_length=False,
                        display_col_length=False, entire_line=False, suggestion=suggestion)

        log(self, (msg, ))
        super().__init__(msg)


class MinecraftSyntaxWarning(SyntaxError):
    """Warnings about dubious Minecraft syntax"""

    def __init__(self, message: str, token: "Token", tokenizer: "Tokenizer", *, col_length: bool = False,
                 display_col_length: bool = True, entire_line: bool = False, suggestion: str | None = None) -> None:
        msg = error_msg(message, token, tokenizer, col_length,
                        display_col_length, entire_line, suggestion)
        log(self, (msg, ))
        super().__init__(msg)



from collections import defaultdict
from enum import Enum, auto
from json import JSONDecodeError, loads
from typing import Any, Callable

from jmc.compile.utils import convention_jmc_to_mc

from .utils import ArgType, NumberType, find_scoreboard_player_type, verify_args, Arg
from ..datapack import DataPack, Function
from ..exception import JMCDecodeJSONError, JMCMissingValueError, JMCValueError
from ..tokenizer import Token, Tokenizer


class FuncType(Enum):
    BOOL_FUNCTION = auto()
    """Returns a part of `/execute if` command"""
    EXECUTE_EXCLUDED = auto()
    """Cannot be used with `/execute`"""
    JMC_COMMAND = auto()
    """Normal custom JMC function"""
    LOAD_ONCE = auto()
    """Can only be used on load function and used once"""
    LOAD_ONLY = auto()
    """Can only be used on load function"""
    VARIABLE_OPERATION = auto()
    """Returns a minecraft integer to a scoreboard variable"""


class JMCFunction:
    """
    Base function for all custom JMC function
    - Must be decorated by `@func_property`
    - Either call_bool (For BOOL_FUNCTION) or call (for other type) must be implemented

    Function types are
    - `FuncType.BOOL_FUNCTION`: returns a part of `/execute if` command
    - `FuncType.EXECUTE_EXCLUDED`: cannot be used with `/execute`
    - `FuncType.LOAD_ONCE`: can only be used on load function and used once
    - `FuncType.LOAD_ONLY`: can only be used on load function
    - `FuncType.VARIABLE_OPERATION`: returns a minecraft integer to a scoreboard variable
    - `FuncType.JMC_COMMAND`: normal custom JMC function

    :param token: paren_round token containing arguments
    :param datapack: Datapack object
    :param tokenizer: Tokenizer
    :param is_execute: Whether the function is in execute command (Can be None if not used), defaults to None
    :param var: For containing minecraft variable (Can be None if not used), defaults to None

    :raises NotImplementedError: If JMCFunction isn't used with decorator
    :raises NotImplementedError: Missing func_type
    :raises JMCValueError: Missing required positional argument
    :raises NotImplementedError: Call function not implemented
    """
    __slots__ = ('arg_type', 'func_type', 'name',
                 'call_string', 'defaults', '_ignore',
                 'token', 'datapack', 'tokenizer',
                 'is_execute', 'var', 'args',
                 'raw_args', 'number_type')
    _decorated: bool = False
    """A private attribute that will be changed by a decorator to check for missing decorator (Set by decorator)"""
    arg_type: dict[str, ArgType]
    """Dictionary containing all parameter and it's ArgType (Set by decorator)"""
    func_type: FuncType
    """Type of the custom JMC function (Set by decorator)"""
    name: str
    """String for the function's name, for further use (Set by decorator)"""
    call_string: str
    """String that will be used in .jmc to call the function (Set by decorator)"""
    defaults: dict[str, str]
    """Defaults value of each parameter as string (Set by decorator)"""
    number_type: dict[str, NumberType]
    """Dictionary containing some number parameter that need to be specific (Set by decorator)"""
    _ignore: set[str]
    """Private set of parameter for parser to ignore and leave as is (Set by decorator)"""

    token: Token
    """paren_round Token object containing the arguments"""
    datapack: DataPack
    """Datapack object"""
    tokenizer: Tokenizer
    """Current Tokenizer of the token"""
    is_execute: bool | None
    """Whether the function is in execute command (in VARIABLE_OPERATION, JMC_COMMANDS function type)"""
    var: str | None
    """Minecraft scoreboard variable for VARIABLE_OPERATION function to return"""

    args: dict[str, str]
    """Dictionary containing parameter and parsed argument in form of string"""
    raw_args: dict[str, Arg]
    """Dictionary containing parameter and given argument as Arg object"""

    __subcls: dict[FuncType, dict[str, type["JMCFunction"]]
                   ] = defaultdict(dict)
    """:cvar: Dictionary of (Function type and according dictionary of funtion name and a subclass)"""

    def __new__(cls, *args, **kwargs):
        if cls is JMCFunction:
            raise TypeError(
                f"Only children of '{cls.__name__}' may be instantiated")
        return super().__new__(cls)

    def __init__(self, token: Token, datapack: DataPack, tokenizer: Tokenizer,
                 *, is_execute: bool | None = None, var: str | None = None) -> None:
        self.token = token
        self.datapack = datapack
        self.tokenizer = tokenizer
        self.is_execute = is_execute
        self.var = var
        if not self._decorated:
            raise NotImplementedError("Missing decorator")
        if self.func_type is None:
            raise NotImplementedError("Missing func_type")

        args_Args = verify_args(self.arg_type,
                                self.call_string, token, tokenizer)
        self.args = {}
        self.raw_args = {}

        for key, arg in args_Args.items():
            if arg is None:
                if key not in self.defaults:
                    raise JMCMissingValueError(key, token, tokenizer)
                self.args[key] = self.defaults[key]
                continue

            self.raw_args[key] = arg
            if key in self._ignore:
                pass
            elif arg.arg_type == ArgType._FUNC_CALL:
                if ':' in arg.token.string:
                    self.args[key] = f"function {arg.token.string}"
                else:
                    self.args[
                        key] = f"function {datapack.namespace}:{convention_jmc_to_mc(arg.token, self.tokenizer)}"
            elif arg.arg_type == ArgType.ARROW_FUNC:
                self.args[key] = '\n'.join(
                    datapack.parse_function_token(arg.token, tokenizer))
            elif arg.arg_type == ArgType.INTEGER:
                self.args[key] = arg.token.string
                # self.args[key] = str(find_scoreboard_player_type(
                #     arg.token, tokenizer).value)
            elif arg.arg_type == ArgType.FLOAT:
                self.args[key] = str(float(arg.token.string))
            elif arg.arg_type in {ArgType.SCOREBOARD_PLAYER, ArgType.SCOREBOARD}:
                scoreboard_player = find_scoreboard_player_type(
                    arg.token, tokenizer)
                if isinstance(scoreboard_player.value, int):
                    raise ValueError(
                        "scoreboard_player.value is int for minecraft scorboard")
                self.args[key] = f"{scoreboard_player.value[1]} {scoreboard_player.value[0]}"
            else:
                self.args[key] = arg.token.string

        for parameter, number_type in self.number_type.items():
            if self.arg_type[parameter] not in {
                    ArgType.INTEGER, ArgType.FLOAT}:
                raise ValueError(f"{parameter} paremeter is not number")

            if number_type == NumberType.POSITIVE:
                if float(self.args[parameter]) <= 0:
                    raise JMCValueError(
                        f"{parameter} can only be {number_type.value}",
                        self.raw_args[parameter].token,
                        tokenizer)
            elif number_type == NumberType.ZERO_POSITIVE:
                if float(self.args[parameter]) < 0:
                    raise JMCValueError(
                        f"{parameter} can only be {number_type.value}",
                        self.raw_args[parameter].token,
                        tokenizer)

            elif number_type == NumberType.NON_ZERO:
                if float(self.args[parameter]) == 0:
                    raise JMCValueError(
                        f"{parameter} can only be {number_type.value}",
                        self.raw_args[parameter].token,
                        tokenizer)

        self.__post__init__()

    def __post__init__(self) -> None:
        """
        This function will be called after initialization of the object
        """

    def call(self) -> str:
        """
        This function will be called when user call matching JMC custom function

        :raises NotImplementedError: When the subclass's call method is not implemented
        :return: Minecraft command as string
        """
        raise NotImplementedError("Call function not implemented")

    def call_bool(self) -> tuple[str, bool]:
        """
        This function will be called when user call matching JMC boolean(Can only be used in condition) function

        :raises NotImplementedError: When the subclass's call method is not implemented
        :return: Tuple of Minecraft command as string and boolean(True->if, False->unless)
        """
        raise NotImplementedError("Call(boolean) function not implemented")

    @classmethod
    def get_subclasses(
            cls, func_type: FuncType) -> dict[str, type["JMCFunction"]]:
        """
        Get dictionary of funtion name and a class matching function type

        :param func_type: Function type to search for
        :return: Dictionary of jmcfunction name and jmcfunction class
        """
        if func_type not in cls.__subcls:
            for subcls in cls.__subclasses__():
                cls.__subcls[subcls.func_type][subcls.call_string] = subcls

        return cls.__subcls[func_type]

    def is_never_used(self, call_string: str | None = None) -> bool:
        """
        Add current function to datapack.used_command and return whether it's already there

        :param: Any string instead of default call string for more specific check. For example, minecraft command
        :return: Whether this function has been called by the user before
        """
        if call_string is None:
            call_string = self.call_string
        is_in = call_string not in self.datapack.used_command
        self.datapack.used_command.add(call_string)
        return is_in

    def get_private_function(self, function_name: str) -> Function:
        """
        Get private function from self.datapack.private_functions

        :param function_name: Name of the private function
        :return: Function object
        """
        return self.datapack.private_functions[self.name][function_name]

    def make_empty_private_function(self, function_name: str) -> Function:
        """
        Make private function with no content

        :param function_name: Name of the function
        :return: Function object
        """
        func = self.datapack.private_functions[
            self.name][function_name] = Function()
        return func

    def load_arg_json(self, parameter: str) -> dict[str, Any]:
        """
        Get JSON argument from parameter name

        :param parameter: Name of the parameter of JMC function
        :raises JMCDecodeJSONError: Invalid JSON
        :return: JSON
        """
        try:
            json = loads(self.args[parameter])
        except JSONDecodeError as error:
            raise JMCDecodeJSONError(
                error, self.raw_args[parameter].token, self.tokenizer) from error

        return json

    def check_bool(self, parameter: str) -> bool:
        """
        Check if argument is true or false

        :param parameter: String parameter
        :raises JMCValueError: Argument is invalid
        :return: Whether argument is true
        """
        if self.args[parameter] not in {"true", "false"}:
            raise JMCValueError(
                f"'{parameter}' only accepts true or false in {self.call_string}",
                self.raw_args[parameter].token,
                self.tokenizer)
        return self.args[parameter] == "true"


def func_property(func_type: FuncType, call_string: str, name: str, arg_type: dict[str, ArgType], defaults: dict[str, str] = {
}, ignore: set[str] = set(), number_type: dict[str, NumberType] = {}) -> Callable[[type[JMCFunction]], type[JMCFunction]]:
    """
    Decorator factory for setting property of custom JMC function

    :param func_type: Type of the custom JMC function
    :param call_string: String that will be used in .jmc to call the function
    :param name: String for the function's name, for further use
    :param arg_type: Dictionary containing all parameter and it's ArgType
    :param defaults: Defaults value of each parameter as string, defaults to {}
    :param ignore: Set of parameter for parser to ignore and leave as is, defaults to set()
    :param number_type: Dictionary containing some number parameter that need to be specific, defaults to {}
    :return: A decorator for JMCFunction class
    """
    def decorator(cls: type[JMCFunction]) -> type[JMCFunction]:
        """
        A decorator to set the class's attributes for setting JMC function's properties

        :param cls: A subclass of JMCFunction
        :return: Same class that was passed in
        """
        cls.func_type = func_type
        cls.call_string = call_string
        cls.arg_type = arg_type
        cls.defaults = defaults
        cls.number_type = number_type
        # for default in defaults:
        #     if default not in arg_type:
        #         raise BaseException()
        cls._ignore = ignore
        cls.name = name

        cls._decorated = True
        return cls
    return decorator



import os
from .configuration import GlobalData
from .utils import Colors, get_input, pprint
from ..compile import Logger

logger = Logger(__name__)

global_data = GlobalData()


def unknown_command(*args, **kwargs) -> None:
    pprint("Command not recognized, try `help` for more info.", Colors.FAIL)


def handle_command(given_command: str) -> None:
    if not given_command:
        return
    command_name, *arguments = given_command.split()
    command_func, usage = global_data.commands.get(
        command_name, (unknown_command, ""))
    try:
        command_func(*arguments)
    except TypeError as error:
        msg: str = error.args[0]
        msg = msg.replace(
            "()",
            " command").replace(
            "positional argument",
            "argument")
        pprint(msg, Colors.FAIL)
        pprint(f"Usage: {usage}", Colors.FAIL)


def start() -> None:
    os.system("")
    logger.info(f"Build-version: {global_data.VERSION}")
    pprint(' JMC Compiler\n', Colors.HEADER)
    pprint(f'Current Directory | {global_data.cwd}\n', Colors.YELLOW)
    if not global_data.config.is_file_exist():
        global_data.config.ask_config()
        global_data.config.save_config()
    else:
        global_data.config.load_config()
    pprint("To compile, type `compile`. For help, type `help`", Colors.INFO)
    while True:
        handle_command(get_input())



"""Module containing JMCFunction subclasses for custom JMC function"""

import math
from typing import Iterator
from ...exception import JMCSyntaxException, JMCValueError
from ..utils import ArgType, NumberType
from ..jmc_function import JMCFunction, FuncType, func_property


def drange(start: float | int, stop: float | int,
           step: float | int) -> Iterator[float | int]:
    """
    Similar to built-in range() function but with float

    :param start: Included starting point
    :param stop: Excluded ending point
    :param step: Step
    :yield: Float/Integer
    """
    result = start
    while result < stop:
        yield result
        result += step


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Timer.set',
    arg_type={
        "objective": ArgType.KEYWORD,
        "selector": ArgType.SELECTOR,
        "tick": ArgType.SCOREBOARD_PLAYER
    },
    name='timer_set'
)
class TimerSet(JMCFunction):
    """
    ```py
    TimerSet(objective, target_selector: selector, tick: number|$variable)
    ```
    Set entity's score to start the timer.
    """

    def call(self) -> str:
        if self.raw_args["tick"].arg_type == ArgType.INTEGER:
            return f'scoreboard players set {self.args["selector"]} {self.args["objective"]} {self.args["tick"]}'
        return f'scoreboard players operations {self.args["selector"]} {self.args["objective"]} = {self.args["tick"]}'


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Item.give',
    arg_type={
        "itemId": ArgType.KEYWORD,
        "selector": ArgType.SELECTOR,
        "amount": ArgType.INTEGER
    },
    name='item_give',
    defaults={
        "selector": "@s",
        "amount": "1"
    },
    number_type={
        "amount": NumberType.ZERO_POSITIVE
    }
)
class ItemGive(JMCFunction):
    def call(self) -> str:
        if self.args["itemId"] not in self.datapack.data.item:
            raise JMCValueError(
                f'Item id: \'{self.args["itemId"]}\' is not defined.',
                self.token,
                self.tokenizer,
                suggestion=f"Use Item.create to make this item BEFORE using {self.call_string}"
            )
        return f'give {self.args["selector"]} {self.datapack.data.item[self.args["itemId"]]} {self.args["amount"]}'


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Item.summon',
    arg_type={
        "itemId": ArgType.KEYWORD,
        "pos": ArgType.STRING,
        "count": ArgType.INTEGER,
        "nbt": ArgType.JS_OBJECT
    },
    name='item_summon',
    defaults={
        "pos": "~ ~ ~",
        "count": "1",
        "nbt": ""
    },
    number_type={
        "count": NumberType.POSITIVE
    }
)
class ItemSummon(JMCFunction):
    def call(self) -> str:
        if self.args["itemId"] not in self.datapack.data.item:
            raise JMCValueError(
                f'Item id: \'{self.args["itemId"]}\' is not defined.',
                self.token,
                self.tokenizer,
                suggestion=f"Use Item.create to make this item BEFORE using {self.call_string}"
            )
        nbt = self.tokenizer.parse_js_obj(
            self.raw_args["nbt"].token) if self.args["nbt"] else {}
        if "Item" in nbt:
            raise JMCValueError(
                f'`Item` key found inside Item.summon nbt argument.',
                self.token,
                self.tokenizer,
                suggestion=f"Remove `Item` in the nbt"
            )
        return f'summon item {self.args["pos"]} {{Item:{{id:"{self.datapack.data.item[self.args["itemId"]].item_type}",Count:{self.args["count"]},tag:{self.datapack.data.item[self.args["itemId"]].nbt}}}{","+self.datapack.token_dict_to_raw_js_object(nbt, self.tokenizer)[1:-1] if nbt else ""}}}'


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Item.replaceBlock',
    arg_type={
        "itemId": ArgType.KEYWORD,
        "pos": ArgType.STRING,
        "slot": ArgType.STRING,
        "count": ArgType.INTEGER
    },
    name='item_replace_block',
    defaults={
        "count": "1"
    },
    number_type={
        "count": NumberType.POSITIVE
    }
)
class ItemReplaceBlock(JMCFunction):
    def call(self) -> str:
        if self.args["itemId"] not in self.datapack.data.item:
            raise JMCValueError(
                f'Item id: \'{self.args["itemId"]}\' is not defined.',
                self.token,
                self.tokenizer,
                suggestion=f"Use Item.create to make this item BEFORE using {self.call_string}"
            )
        return f'item replace block {self.args["pos"]} {self.args["slot"]} with {self.datapack.data.item[self.args["itemId"]]} {self.args["count"]}'


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Item.replaceEntity',
    arg_type={
        "itemId": ArgType.KEYWORD,
        "selector": ArgType.SELECTOR,
        "slot": ArgType.STRING,
        "count": ArgType.INTEGER
    },
    name='item_replace_entity',
    defaults={
        "count": "1"
    },
    number_type={
        "count": NumberType.POSITIVE
    }
)
class ItemReplaceEntity(JMCFunction):
    def call(self) -> str:
        if self.args["itemId"] not in self.datapack.data.item:
            raise JMCValueError(
                f'Item id: \'{self.args["itemId"]}\' is not defined.',
                self.token,
                self.tokenizer,
                suggestion=f"Use Item.create to make this item BEFORE using {self.call_string}"
            )
        return f'item replace entity {self.args["selector"]} {self.args["slot"]} with {self.datapack.data.item[self.args["itemId"]]} {self.args["count"]}'


def __normalize_decimal(n: float):
    """
    Parse float into string

    :param n: _description_
    """
    if n == 0:
        return ""
    stripped = f'{n:.10f}'  # .rstrip('0').rstrip('.')
    return stripped


def points_to_commands(points: list[tuple[float, float, float]], particle: str,
                       speed: str, count: str, mode: str, notation: str = "^") -> list[str]:
    """
    Parse list of points(x,y,z position) into particle commands

    :param points: List of points(x,y,z position)
    :param particle: particle type
    :param speed: particle speed
    :param count: particle count
    :param mode: particle mode
    :param notation: Notation (~) or (^)
    :return: particle commands
    """
    commands = []
    for x_pos, y_pos, z_pos in points:
        commands.append(
            f'particle {particle} {notation}{__normalize_decimal(x_pos)} {notation}{__normalize_decimal(y_pos)} {notation}{__normalize_decimal(z_pos)} 0 0 0 {speed} {count} {mode}')
    return commands


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Particle.circle',
    arg_type={
        "particle": ArgType.STRING,
        "radius": ArgType.FLOAT,
        "spread": ArgType.INTEGER,
        "speed": ArgType.INTEGER,
        "count": ArgType.INTEGER,
        "mode": ArgType.KEYWORD,
    },
    name='particle_circle',
    defaults={
        "speed": "1",
        "count": "1",
        "mode": "normal",
    },
    number_type={
        "spread": NumberType.POSITIVE,
        "radius": NumberType.POSITIVE,
        "count": NumberType.POSITIVE,
        "speed": NumberType.ZERO_POSITIVE
    }
)
class ParticleCircle(JMCFunction):
    def draw(self, radius: float,
             spread: int) -> list[tuple[float, float, float]]:
        points = []
        angle = 2 * math.pi / spread
        for i in drange(0, spread, angle):
            points.append((radius * math.cos(i), 0.0, radius * math.sin(i)))
        return points

    def call(self) -> str:
        if self.args['mode'] not in {'force', 'normal'}:
            raise JMCSyntaxException(
                f"Unrecognized mode, '{self.args['mode']}' Available modes are 'force' and 'normal'", self.raw_args["mode"].token, self.tokenizer)

        return self.datapack.add_raw_private_function(
            self.name,
            commands=points_to_commands(
                self.draw(
                    float(self.args["radius"]),
                    int(self.args["spread"])),
                self.args["particle"],
                self.args["speed"],
                self.args["count"],
                self.args["mode"]
            ),
        )


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Particle.spiral',
    arg_type={
        "particle": ArgType.STRING,
        "radius": ArgType.FLOAT,
        "height": ArgType.FLOAT,
        "spread": ArgType.INTEGER,
        "speed": ArgType.INTEGER,
        "count": ArgType.INTEGER,
        "mode": ArgType.KEYWORD,
    },
    name='particle_spiral',
    defaults={
        "speed": "1",
        "count": "1",
        "mode": "normal",
    },
    number_type={
        "spread": NumberType.POSITIVE,
        "radius": NumberType.POSITIVE,
        "height": NumberType.POSITIVE,
        "speed": NumberType.ZERO_POSITIVE
    }
)
class ParticleSpiral(JMCFunction):
    def draw(self, radius: float, height: float,
             spread: int) -> list[tuple[float, float, float]]:
        points = []
        angle = 2 * math.pi / spread
        d_y = height / spread
        for i in range(spread):
            points.append((
                radius * math.cos(i * angle),
                i * d_y,
                radius * math.sin(i * angle)))
        return points

    def call(self) -> str:
        if self.args['mode'] not in {'force', 'normal'}:
            raise JMCSyntaxException(
                f"Unrecognized mode, '{self.args['mode']}' Available modes are 'force' and 'normal'", self.raw_args["mode"].token, self.tokenizer)

        return self.datapack.add_raw_private_function(
            self.name,
            commands=points_to_commands(
                self.draw(
                    float(self.args["radius"]),
                    float(self.args["height"]),
                    int(self.args["spread"])),
                self.args["particle"],
                self.args["speed"],
                self.args["count"],
                self.args["mode"]
            ),
        )


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Particle.cylinder',
    arg_type={
        "particle": ArgType.STRING,
        "radius": ArgType.FLOAT,
        "height": ArgType.FLOAT,
        "spreadXZ": ArgType.INTEGER,
        "spreadY": ArgType.INTEGER,
        "speed": ArgType.INTEGER,
        "count": ArgType.INTEGER,
        "mode": ArgType.KEYWORD,
    },
    name='particle_cylinder',
    defaults={
        "speed": "1",
        "count": "1",
        "mode": "normal",
    },
    number_type={
        "spreadXZ": NumberType.POSITIVE,
        "spreadY": NumberType.POSITIVE,
        "radius": NumberType.POSITIVE,
        "height": NumberType.POSITIVE,
        "count": NumberType.POSITIVE,
        "speed": NumberType.ZERO_POSITIVE
    }
)
class ParticleCylinder(JMCFunction):
    def draw(self, radius: float, height: float, spread_xz: int,
             spread_y: int) -> list[tuple[float, float, float]]:
        points = []
        d_y = height / spread_y
        for y in range(spread_y):
            angle = 2 * math.pi / spread_xz
            for i in drange(0, spread_xz, angle):
                points.append(
                    (radius * math.cos(i), y * d_y, radius * math.sin(i)))
        return points

    def call(self) -> str:
        if self.args['mode'] not in {'force', 'normal'}:
            raise JMCSyntaxException(
                f"Unrecognized mode, '{self.args['mode']}' Available modes are 'force' and 'normal'", self.raw_args["mode"].token, self.tokenizer)

        return self.datapack.add_raw_private_function(
            self.name,
            commands=points_to_commands(
                self.draw(
                    float(self.args["radius"]),
                    float(self.args["height"]),
                    int(self.args["spreadXZ"]),
                    int(self.args["spreadY"])),
                self.args["particle"],
                self.args["speed"],
                self.args["count"],
                self.args["mode"]
            ),
        )


@func_property(
    func_type=FuncType.JMC_COMMAND,
    call_string='Particle.line',
    arg_type={
        "particle": ArgType.STRING,
        "distance": ArgType.FLOAT,
        "spread": ArgType.INTEGER,
        "speed": ArgType.INTEGER,
        "count": ArgType.INTEGER,
        "mode": ArgType.KEYWORD,
    },
    name='particle_line',
    defaults={
        "speed": "1",
        "count": "1",
        "mode": "normal",
    },
    number_type={
        "spread": NumberType.POSITIVE,
        "distance": NumberType.POSITIVE,
        "count": NumberType.POSITIVE,
        "speed": NumberType.ZERO_POSITIVE
    }
)
class ParticleLine(JMCFunction):
    def draw(self, distance: float,
             spread: int) -> list[tuple[float, float, float]]:
        return [(0, 0, n) for n in drange(1, distance + 1, distance / spread)]

    def call(self) -> str:
        if self.args['mode'] not in {'force', 'normal'}:
            raise JMCSyntaxException(
                f"Unrecognized mode, '{self.args['mode']}' Available modes are 'force' and 'normal'", self.raw_args["mode"].token, self.tokenizer)

        return self.datapack.add_raw_private_function(
            self.name,
            commands=points_to_commands(
                self.draw(
                    float(self.args["distance"]),
                    int(self.args["spread"])),
                self.args["particle"],
                self.args["speed"],
                self.args["count"],
                self.args["mode"]
            ),
        )



"""Module for testing compilation"""
from json import dumps
from pathlib import Path

from ..terminal.configuration import Configuration, GlobalData
from .log import Logger
from .header import Header
from .compiling import read_cert, read_header, build
from .lexer import Lexer


logger = Logger(__name__)


class JMCPack:
    """
    Class representation of folder structure containing entire JMC files

    - Used for testing to prevent having to write and read files

    :param namespace: Namespace of the virtual datapack, defaults to "TEST"
    :param output: virtual directory for output, defaults to "VIRTUAL"
    """
    __slots__ = ('cert', 'jmc_file', 'header_file',
                 '__built', 'config')
    cert: str
    jmc_file: str
    header_file: str | None
    __built: dict[str, str] | None
    """Dictionary of file name and file content"""

    def __init__(self, namespace: str = "TEST",
                 output: str = "VIRTUAL") -> None:
        self.__built = None
        self.cert = """LOAD=__load__
TICK=__tick__
PRIVATE=__private__
VAR=__variable__
INT=__int__"""
        self.jmc_file = ""
        self.header_file = None
        self.config = Configuration(
            GlobalData(),
            namespace=namespace,
            description="__THIS_IS_FOR_TESTING__",
            pack_format="10",
            target=Path(__file__),
            output=Path(output)
        )

    def set_cert(self, file_content: str) -> "JMCPack":
        """
        Set certificate(jmc.txt)

        :param file_content: File's content
        :return: Self
        """
        self.cert = file_content
        return self

    def set_jmc_file(self, file_content: str) -> "JMCPack":
        """
        Set main jmc file

        :param file_content: File's content
        :return: Self
        """
        self.jmc_file = file_content
        return self

    def set_header_file(self, file_content: str) -> "JMCPack":
        """
        Set main header file

        :param file_content: File's content
        :return: Self
        """
        self.header_file = file_content
        return self

    def build(self) -> "JMCPack":
        """
        Build datapack
        :return: Self
        """
        logger.info("Building from JMCPack")
        Header.clear()
        read_cert(self.config, _test_file=self.cert)
        read_header(self.config, _test_file=self.header_file)
        lexer = Lexer(self.config, _test_file=self.jmc_file)
        self.__built = build(lexer.datapack, self.config, _is_virtual=True)
        return self

    @property
    def built(self) -> dict[str, str]:
        if self.__built is None:
            self.build()
        if self.__built is None:
            raise ValueError("self.__built is None")
        return self.__built

    def dumps(self, *, indent: int = 4) -> str:
        return dumps(self.built, indent=indent)

    def to_string(self) -> str:
        return "\n".join(
            [f"> {file_name}\n{file_content}" for file_name, file_content in self.built.items()])

    def __str__(self) -> str:
        ouput_str = ""
        for key, value in self.built.items():
            ouput_str += f"""
=============================
{key}
-----------------------------
{value}
=============================
"""
        return ouput_str



"""Module containing vanilla commands"""
COMMANDS = {
    # OP Level 0-2
    "advancement",
    "attribute",
    "bossbar",
    "clear",
    "clone",
    "data",
    "datapack",
    "debug",
    "defaultgamemode",
    "difficulty",
    "effect",
    "enchant",
    "execute",
    "experience",
    "fill",
    "forceload",
    "function",
    "gamemode",
    "gamerule",
    "give",
    "help",
    "item",
    "kill",
    "list",
    "locate",
    "locatebiome",
    "loot",
    "me",
    "msg",
    "particle",
    "placefeature",
    "playsound",
    "recipe",
    "reload",
    "say",
    "schedule",
    "scoreboard",
    "seed",
    "setblock",
    "setworldspawn",
    "spawnpoint",
    "spectate",
    "spreadplayers",
    "stopsound",
    "summon",
    "tag",
    "team",
    "teammsg",
    "teleport",
    "tell",
    "tellraw",
    "time",
    "title",
    "tm",
    "tp",
    "trigger",
    "w",
    "weather",
    "whitelist",
    "worldborder",
    "xp"
    # OP Level 3-4
    "jfr",
    "perf",
    "publish",
    "save-all",
    "save-off",
    "save-on",
    "stop",
    "ban",
    "ban-ip",
    "banlist",
    "debug",
    "deop",
    "kick",
    "op",
    "pardon",
    "pardon-ip",
    "setidletimeout",
    "whitelist"
}
"""All vanilla command (first argument only)"""



"""Untility for commands"""
import ast
import json
import operator as op
import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Callable

from ..datapack import DataPack
from ..tokenizer import Token, Tokenizer, TokenType
from ..exception import JMCSyntaxException, JMCValueError
from ..utils import is_number, is_float


class PlayerType(Enum):
    VARIABLE = auto()
    INTEGER = auto()
    SCOREBOARD = auto()


@dataclass(frozen=True, slots=True)
class ScoreboardPlayer:
    player_type: PlayerType
    value: int | tuple[str, str]
    """Contains either integer or (objective and selector)"""


def find_scoreboard_player_type(
        token: Token, tokenizer: Tokenizer, allow_integer: bool = True) -> ScoreboardPlayer:
    """
    Generate ScoreboardPlayer including its type from a keyword token

    :param token: keyword token to parse
    :param tokenizer: token's Tokenizer
    :param allow_integer: Whether to allow integer(rvalue), defaults to True
    :raises JMCSyntaxException: Token is not a keyword token
    :return: ScoreboardPlayer
    """
    if token.token_type != TokenType.KEYWORD:
        raise JMCSyntaxException(
            "Expected keyword", token, tokenizer)

    if token.string.startswith(DataPack.VARIABLE_SIGN):
        return ScoreboardPlayer(player_type=PlayerType.VARIABLE, value=(
            DataPack.var_name, token.string))

    if is_number(token.string):
        return ScoreboardPlayer(
            player_type=PlayerType.INTEGER, value=int(token.string))

    splits = token.string.split(':')
    if len(splits) == 1:
        if allow_integer:
            raise JMCSyntaxException(
                "Expected integer, variable, or objective:selector", token, tokenizer)
        raise JMCSyntaxException(
            "Expected variable or objective:selector", token, tokenizer)
    if len(splits) > 2:
        raise JMCSyntaxException(
            "Scoreboard's player cannot contain more than 1 colon(:)", token, tokenizer)

    return ScoreboardPlayer(
        player_type=PlayerType.SCOREBOARD, value=(splits[0], splits[1]))


class NumberType(Enum):
    POSITIVE = "more than zero"
    ZERO_POSITIVE = "more than or equal to zero"
    NON_ZERO = "non-zero"


class ArgType(Enum):
    ARROW_FUNC = "arrow(anonymous) function"
    JS_OBJECT = "JavaScript object (dictionary)"
    JSON = "JSON"
    SCOREBOARD = "variable or objective:selector"
    INTEGER = "integer"
    FLOAT = "integer or decimal(real number)"
    STRING = "string"
    KEYWORD = "keyword"
    SELECTOR = "target selector"
    LIST = "list/array"
    FUNC = "function"
    _FUNC_CALL = "function"
    SCOREBOARD_PLAYER = "integer, variable, or objective:selector"
    ANY = None


class Arg:
    __slots__ = ('token', 'arg_type')

    def __init__(self, token: Token, arg_type: ArgType) -> None:
        self.token = token
        self.arg_type = arg_type

    def verify(self, verifier: ArgType, tokenizer: Tokenizer,
               key_string: str) -> "Arg":
        """
        Verify if the argument is valid

        :param verifier: Argument's type
        :param tokenizer: _description_
        :param key_string: Key(kwarg) in from of string
        :return: self
        """
        if verifier == ArgType.ANY:
            return self
        if verifier == ArgType.SCOREBOARD_PLAYER:
            if self.arg_type in {ArgType.SCOREBOARD, ArgType.INTEGER}:
                return self
            raise JMCValueError(
                f"For '{key_string}' key, expected {verifier.value}, got {self.arg_type.value}", self.token, tokenizer)
        if verifier == ArgType.FLOAT:
            if self.arg_type in {ArgType.FLOAT, ArgType.INTEGER}:
                return self
            raise JMCValueError(
                f"For '{key_string}' key, expected {verifier.value}, got {self.arg_type.value}", self.token, tokenizer)
        if verifier == ArgType.FUNC:
            if self.arg_type == ArgType.ARROW_FUNC:
                return self
            if self.arg_type in {ArgType.KEYWORD, ArgType.SCOREBOARD}:
                self.arg_type = ArgType._FUNC_CALL
                return self
            raise JMCValueError(
                f"For '{key_string}' key, expected {verifier.value}, got {self.arg_type.value}", self.token, tokenizer)
        if verifier == ArgType.KEYWORD:
            if key_string.startswith('@'):
                raise JMCValueError(
                    f"For '{key_string}' key, expected {verifier.value}, got {ArgType.SELECTOR.value}",
                    self.token,
                    tokenizer)
            if ':' in key_string:
                raise JMCValueError(
                    f"For '{key_string}' key, expected {verifier.value}, got {ArgType.SCOREBOARD.value}",
                    self.token,
                    tokenizer)
        if verifier != self.arg_type:
            suggestion = None
            if verifier == ArgType.STRING:
                suggestion = f"Did you mean: ' \"{self.token.string}\" '"
            raise JMCValueError(
                f"For '{key_string}' key, expected {verifier.value}, got {self.arg_type.value}", self.token, tokenizer, suggestion=suggestion)
        return self


def find_arg_type(token: Token, tokenizer: Tokenizer) -> ArgType:
    """
    Find type of the argument in form of token and return the ArgType

    :param token: any token representing argument
    :param tokenizer: Tokenizer
    :raises JMCValueError: Cannot find ArgType.FUNC type
    :return: Argument's type of the token
    """
    if token.token_type == TokenType.FUNC:
        return ArgType.ARROW_FUNC
    if token.token_type == TokenType.PAREN_CURLY:
        if re.match(r'^{\s*"', token.string) is not None:
            return ArgType.JSON
        return ArgType.JS_OBJECT
    if token.token_type == TokenType.PAREN_SQUARE:
        return ArgType.LIST

    if token.token_type == TokenType.KEYWORD:
        if token.string.startswith(
                DataPack.VARIABLE_SIGN) or ':' in token.string:
            return ArgType.SCOREBOARD
        if is_number(token.string):
            return ArgType.INTEGER
        if is_float(token.string):
            return ArgType.FLOAT
        if token.string.startswith('@'):
            return ArgType.SELECTOR
        return ArgType.KEYWORD
    if token.token_type == TokenType.STRING:
        return ArgType.STRING

    raise JMCValueError(
        "Unknown argument type", token, tokenizer)


def verify_args(params: dict[str, ArgType], feature_name: str,
                token: Token, tokenizer: Tokenizer) -> dict[str, Arg | None]:
    """
    Verify argument types of a paren_round token

    :param params: Dictionary of arguments(string) and its type(ArgType)
    :param feature_name: Feature name to show up in error
    :param token: paren_round token
    :param tokenizer: token's tokenizer
    :raises JMCValueError: Got too many positional arguments
    :raises JMCValueError: Unknown key
    :return: Dictionary of arguments(string) and said argument in Arg form
    """
    args, kwargs = tokenizer.parse_func_args(token)
    result: dict[str, Arg | None] = {key: None for key in params}
    key_list = list(params)
    if len(args) > len(key_list):
        raise JMCValueError(
            f"{feature_name} takes {len(key_list)} positional arguments, got {len(args)}", token, tokenizer)
    for key, arg in zip(key_list, args):
        arg_type = find_arg_type(arg, tokenizer)
        result[key] = Arg(arg, arg_type).verify(params[key], tokenizer, key)
    for key, kwarg in kwargs.items():
        if key not in key_list:
            raise JMCValueError(
                f"{feature_name} got unexpected keyword argument '{key}'", kwarg, tokenizer,
                suggestion=f"""Available arguments are\n{', '.join(f"'{param}'" for param in params)}""")
        arg_type = find_arg_type(kwarg, tokenizer)
        result[key] = Arg(kwarg, arg_type).verify(params[key], tokenizer, key)
    return result


def eval_expr(expr: str) -> str:
    """
    Evaluate mathematical expression and calculate the result number then cast it to string

    :param expr: Expression string
    :return: String representation of result number
    """
    return str(__eval(ast.parse(expr, mode='eval').body))


OPERATORS: dict[type, Callable[..., Any]] = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
                                             ast.Div: op.truediv, ast.Pow: op.pow,
                                             ast.USub: op.neg}


def __eval(node):
    """
    Inner working of eval_expr

    :param node: expr(body of Expression returned from ast.parse)
    :raises TypeError: Invalid type of node
    :return: Result number
    """
    if isinstance(node, ast.Num):  # <number>
        return node.n
    if isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return OPERATORS[type(node.op)](__eval(node.left), __eval(node.right))
    if isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return OPERATORS[type(node.op)](__eval(node.operand))

    raise TypeError(node)


class FormattedText:
    """
    Parse formatted text into raw json string

    :param raw_text: Text
    :param token: Token
    :param tokenizer: Tokenizer
    :param is_default_no_italic: Whether to set italic to False by default, defaults to False
    :return: Raw JSON string

    .. Example::
    >>> minecraft_formatted_text("&4&lName")
    '{"text":"Name", "color":"red", "bold":true}'
    """
    __slots__ = ('raw_text', 'current_json', 'result',
                 'bracket_content', 'token', 'tokenizer',
                 'current_color', 'is_default_no_italic')

    SIGN = "&"
    OPEN_BRACKET = "<"
    CLOSE_BRACKET = ">"
    PROPS = {
        "1": "dark_blue",
        "2": "dark_green",
        "3": "dark_aqua",
        "4": "dark_red",
        "5": "dark_purple",
        "6": "gold",
        "7": "gray",
        "8": "dark_gray",
        "9": "blue",
        "0": "black",
        "a": "green",
        "b": "aqua",
        "c": "red",
        "d": "light_purple",
        "e": "yellow",
        "f": "white",
        "k": "obfuscated",
        "l": "bold",
        "m": "strikethrough",
        "n": "underlined",
        "o": "italic",
        "r": "reset",
    }

    def __init__(self, raw_text: str, token: Token,
                 tokenizer: Tokenizer, *, is_default_no_italic: bool = False) -> None:
        self.raw_text = raw_text
        self.token = token
        self.tokenizer = tokenizer
        self.is_default_no_italic = is_default_no_italic
        self.bracket_content = ""
        self.result: list[dict[
            str,
            str | bool | dict[
                str, str | bool
            ]
        ]] = []
        self.current_json: dict[str, str | bool] = {"text": ""}
        self.current_color = ""
        self.parse()

    def add_key(self, key: str, value: str | bool |
                dict[str, str | bool]) -> None:
        if self.result:
            self.result[0][key] = value
            return

        self.result.append({key: value})

    def push(self) -> None:
        """
        Append current_json to result and reset it
        """
        if not self.current_json["text"]:
            return
        self.result.append(self.current_json)  # type: ignore
        self.current_json = {"text": ""}

    def parse_bracket(self) -> None:
        """
        Parse self.bracket_content and reset it
        """
        bracket_content = self.bracket_content
        self.bracket_content = ""
        properties = bracket_content.split(",")
        for prop in properties:
            value = True
            prop = prop.strip()
            if prop.startswith("!"):
                value = False
                prop = prop[1:]
            if prop in {
                'dark_red',
                'red',
                'gold',
                'yellow',
                'dark_green',
                'green',
                'aqua',
                'dark_aqua',
                'blue',
                'light_purple',
                'dark_purple',
                'white',
                'gray',
                'dark_gray',
                'black',
                'reset',
            }:
                if 'color' in self.current_json:
                    raise JMCValueError(
                        f'color({prop}) used twice in formatted text', self.token, self.tokenizer)

                self.current_json['color'] = prop
                self.current_color = prop
                if not value:
                    raise JMCValueError(
                        f'Color({prop}) cannot be false', self.token, self.tokenizer)
                continue

            if prop in {'bold', 'italic', 'underlined',
                        'strikethrough', 'obfuscated'}:
                self.current_json[prop] = value
                continue

            if prop.startswith('#') and len(prop) == 7:
                if 'color' in self.current_json:
                    raise JMCValueError(
                        f'color({prop}) used twice in formatted text', self.token, self.tokenizer)

                self.current_json['color'] = prop
                if not value:
                    raise JMCValueError(
                        f'Color({prop}) cannot be false', self.token, self.tokenizer)
                continue

            raise JMCValueError(
                f"Unknown property '{prop}'", self.token, self.tokenizer)

        if 'color' not in self.current_json and self.current_color:
            self.current_json['color'] = self.current_color

    def parse_code(self, char: str) -> None:
        """
        Parse color code

        :param char: A character in color code
        """
        prop = self.PROPS.get(char, None)
        if prop is None:
            JMCValueError(
                f"Unknown code format '{char}'", self.token, self.tokenizer)

        if prop in {
            'dark_red',
            'red',
            'gold',
            'yellow',
            'dark_green',
            'green',
            'aqua',
            'dark_aqua',
            'blue',
            'light_purple',
            'dark_purple',
            'white',
            'gray',
            'dark_gray',
            'black',
            'reset',
        }:
            if 'color' in self.current_json:
                raise JMCValueError(
                    f'color({prop}) used twice in formatted text', self.token, self.tokenizer)

            self.current_json['color'] = prop
            self.current_color = prop

        if prop in {'bold', 'italic', 'underlined',
                    'strikethrough', 'obfuscated'}:
            if self.current_color:
                self.current_json['color'] = self.current_color
            self.current_json[prop] = True

    def parse(self) -> None:
        """
        Parse raw_text
        """
        is_expect_color_code = False
        is_bracket_format = False

        for char in self.raw_text:
            if is_bracket_format:
                if char == self.CLOSE_BRACKET:
                    is_bracket_format = False
                    self.parse_bracket()
                    continue
                self.bracket_content += char
                continue

            if is_expect_color_code:
                is_expect_color_code = False
                if char == self.OPEN_BRACKET:
                    is_bracket_format = True
                    continue
                if char == self.SIGN:
                    if isinstance(self.current_json["text"], str):
                        self.current_json["text"] += char
                        continue
                self.parse_code(char)
                continue

            if char == self.SIGN:
                self.push()
                is_expect_color_code = True
                continue

            if isinstance(self.current_json["text"], str):
                self.current_json["text"] += char

        self.push()

        if is_expect_color_code:
            raise JMCValueError(
                f"Unexpected trailing '{self.SIGN}'", self.token, self.tokenizer, suggestion=f"Remove last '{self.SIGN}'")

    def __str__(self) -> str:
        if len(self.result) == 1:
            if 'italic' not in self.result[0]:
                self.result[0]['italic'] = False
            return json.dumps(self.result[0])

        if not self.result:
            return ''
        return json.dumps(
            [{"italic": False} if self.is_default_no_italic else ""] + self.result)

    def __repr__(self) -> str:
        return f"FormattedText(raw_text={repr(self.raw_text)}, result={repr(json.dumps(self.result))})"



"""Module handling variable operation"""
from .jmc_function import JMCFunction, FuncType
from ..datapack import DataPack
from ..exception import JMCSyntaxException
from ..tokenizer import Token, TokenType, Tokenizer
from .utils import find_scoreboard_player_type, PlayerType

VAR_OPERATION_COMMANDS = JMCFunction.get_subclasses(
    FuncType.VARIABLE_OPERATION)


def variable_operation(
        tokens: list[Token], tokenizer: Tokenizer, datapack: DataPack, is_execute: bool) -> str:
    """
    Parse statement for variable operation including custom JMC command that return and integer to be stored in scoreboard value

    :param tokens: Statement(List of tokens) for parsing
    :param tokenizer: Tokenizer
    :param datapack: Datapack object
    :param is_execute: Whether the statement/function is in `/execute`
    :return: Full minecraft command
    """
    if tokens[0].string.endswith('.get') and len(
            tokens) > 1 and tokens[1].token_type == TokenType.PAREN_ROUND:
        if len(tokens) > 2:
            raise JMCSyntaxException(
                "Unexpected token", tokens[2], tokenizer)

        if tokens[1].string != "()":
            raise JMCSyntaxException(
                "'get' function takes no arguments, expected empty bracket ()", tokens[1], tokenizer)

        return f"scoreboard players get {tokens[0].string[:-4]} {DataPack.var_name}"

    tokens = tokenizer.split_keyword_tokens(
        tokens, ['-', '=', '+', '*', '%', '>', '<'])
    if len(tokens) == 1:
        raise JMCSyntaxException(
            "Expected operator after variable", tokens[0], tokenizer, col_length=True)

    for operator in ('*=', '+=', '-=', '*=', '/=', '%=',
                     '++', '--', '><', "->", '>', '<', '='):  # sort key=len
        list_of_tokens = tokenizer.find_tokens(tokens, operator)
        if len(list_of_tokens) == 1:
            continue

        if len(list_of_tokens) > 2:
            raise JMCSyntaxException(
                f"Duplicated operator({operator})", list_of_tokens[2][-1], tokenizer)

        if len(list_of_tokens[0]) > 1:
            raise JMCSyntaxException(
                f"Unexpected token ('{list_of_tokens[0][1].string}') after variable ('{list_of_tokens[0][0].string}')", list_of_tokens[0][1], tokenizer, suggestion="Expected operator")

        if operator in {'++', '--'}:
            if list_of_tokens[1]:
                raise JMCSyntaxException(
                    f"Unexpected token after '{operator}'", list_of_tokens[1][0], tokenizer)
            if len(list_of_tokens[0]) > 1:
                raise JMCSyntaxException(
                    f"Unexpected token before '{operator}'", list_of_tokens[0][1], tokenizer)

            if operator == '++':
                return f"scoreboard players add {list_of_tokens[0][0].string} {DataPack.var_name} 1"
            if operator == '--':
                return f"scoreboard players remove {list_of_tokens[0][0].string} {DataPack.var_name} 1"

        if not list_of_tokens[1]:
            raise JMCSyntaxException(
                f"Expected token after an operator('{operator}')", tokens[-1], tokenizer, suggestion="Expected integer or variable or target selector")

        if operator == '=' and list_of_tokens[1][0].token_type == TokenType.KEYWORD and list_of_tokens[1][0].string in VAR_OPERATION_COMMANDS:
            if len(list_of_tokens[1]) == 1:
                raise JMCSyntaxException(
                    "Expected (", list_of_tokens[1][0], tokenizer, col_length=True)

            if list_of_tokens[1][1].token_type != TokenType.PAREN_ROUND:
                raise JMCSyntaxException(
                    "Expected (", list_of_tokens[1][1], tokenizer)

            if len(list_of_tokens[1]) > 2:
                raise JMCSyntaxException(
                    "Unexpected token", list_of_tokens[1][2], tokenizer)

            return VAR_OPERATION_COMMANDS[list_of_tokens[1][0].string](
                list_of_tokens[1][1], datapack, tokenizer, var=list_of_tokens[0][0].string, is_execute=is_execute).call()

        if len(list_of_tokens[1]) > 1:
            raise JMCSyntaxException(
                f"Unexpected token ({list_of_tokens[1][1].string})", list_of_tokens[1][1], tokenizer)

        scoreboard_player = find_scoreboard_player_type(
            list_of_tokens[1][0], tokenizer, allow_integer=False)

        if operator == '->':
            if scoreboard_player.player_type == PlayerType.INTEGER:
                raise JMCSyntaxException(
                    "Cannot copy score into integer", list_of_tokens[1][0], tokenizer)
            if isinstance(scoreboard_player.value, int):
                raise ValueError("scoreboard_player.value is int")

            return f"scoreboard players operation {scoreboard_player.value[1]} {scoreboard_player.value[0]} = {list_of_tokens[0][0].string} {DataPack.var_name}"

        if scoreboard_player.player_type == PlayerType.INTEGER:
            if operator == '+=':
                return f"scoreboard players add {list_of_tokens[0][0].string} {DataPack.var_name} {scoreboard_player.value}"
            if operator == '-=':
                return f"scoreboard players remove {list_of_tokens[0][0].string} {DataPack.var_name} {scoreboard_player.value}"
            if operator == '=':
                return f"scoreboard players set {list_of_tokens[0][0].string} {DataPack.var_name} {scoreboard_player.value}"

            if not isinstance(scoreboard_player.value, int):
                raise ValueError("scoreboard_player.value is not int")
            datapack.add_int(scoreboard_player.value)
            return f"scoreboard players operation {list_of_tokens[0][0].string} {DataPack.var_name} {operator} {scoreboard_player.value} {DataPack.int_name}"

        if isinstance(scoreboard_player.value, int):
            raise ValueError("scoreboard_player.value is int")

        return f"scoreboard players operation {list_of_tokens[0][0].string} {DataPack.var_name} {operator} {scoreboard_player.value[1]} {scoreboard_player.value[0]}"

    raise JMCSyntaxException(
        "No operator found in variable operation", tokens[0], tokenizer)



from dataclasses import dataclass


@dataclass(slots=True, frozen=True, eq=True)
class Item:
    item_type: str
    nbt: str

    def __str__(self) -> str:
        return self.item_type + (self.nbt if self.nbt != "" else "")


class Data:
    """
    Data shared across all JMC function in the datapack
    """
    __slots__ = 'item', '__item_id_count', 'condition_count'

    def __init__(self) -> None:
        self.item: dict[str, Item] = {}
        self.__item_id_count: int = 0
        self.condition_count: int = 0

    def get_item_id(self) -> str:
        """
        Get item id for on_click feature of Item.create

        :return: Item ID
        """
        self.__item_id_count += 1
        return str(self.__item_id_count)



"""Module for parsing conditions (statements that return boolean), called from command/_flow_control.py and lexer.py"""

from dataclasses import dataclass
from typing import Union

from ..tokenizer import TokenType, Tokenizer, Token
from ..exception import JMCSyntaxException, JMCValueError
from ..datapack import DataPack
from .utils import find_scoreboard_player_type, PlayerType
from .jmc_function import JMCFunction, FuncType

from .builtin_function import bool_function

AND_OPERATOR = '&&'
OR_OPERATOR = '||'
NOT_OPERATOR = '!'

IF = True
UNLESS = False

VAR = '__logic__'
BOOL_FUNCTIONS = JMCFunction.get_subclasses(FuncType.BOOL_FUNCTION)


@dataclass(eq=False, repr=True, slots=True)
class Condition:
    """
    Dataclass for condition containing string representation(excluding if/unless) and whether it's for `if` or `unless`
    """
    string: str
    if_unless: bool
    """`True` means 'if', `False` means 'unless'"""

    def reverse(self) -> None:
        self.if_unless = not self.if_unless

    def __str__(self) -> str:
        return f"{'if' if self.if_unless else 'unless'} {self.string}"


AST_TYPE = dict[str,  # type: ignore
                      Union[str, list["AST_TYPE"], "AST_TYPE"]  # type: ignore
                ] | Condition


def merge_condition(conditions: list[Condition]) -> str:
    """
    Merge all condition into a single string for minecraft execute if

    :param conditions: List of conditions
    :return: Minecraft arguments after `execute if`
    """
    return ' '.join(str(condition) for condition in conditions)


def custom_condition(
        tokens: list[Token], tokenizer: Tokenizer, datapack: DataPack) -> Condition:
    """
    Create a custom JMC condition from list of tokens representing statement

    :param tokens: List of tokens representing statement for condition
    :param tokenizer: Tokenizer
    :param datapack: Datapack object
    :return: Condition object parsed from list of tokens
    """
    if tokens[0].token_type == TokenType.KEYWORD and tokens[0].string.startswith(
            DataPack.VARIABLE_SIGN):
        tokens = tokenizer.split_keyword_tokens(tokens, ['>', '=', '<', '!'])
        first_token = tokens[0]
        for operator in ('===', '==',
                         '>=', '<=', '!=', '>', '<', '='):  # sort key=len
            list_of_tokens = tokenizer.find_tokens(tokens, operator)
            if len(list_of_tokens) == 1:
                continue
            if len(list_of_tokens) > 2:
                raise JMCSyntaxException(
                    f"Duplicated operator({operator}) in condition", list_of_tokens[2][-1], tokenizer)

            if len(list_of_tokens[0]) > 1:
                raise JMCSyntaxException(
                    f"Unexpected token ('{list_of_tokens[0][1].string}') after variable ('{list_of_tokens[0][0].string}') in condition", list_of_tokens[0][1], tokenizer, suggestion="Expected operator")

            if len(list_of_tokens[1]) > 1:
                raise JMCSyntaxException(
                    "Unexpected token in condition", list_of_tokens[1][-1], tokenizer)

            second_token = list_of_tokens[1][0]
            scoreboard_player = find_scoreboard_player_type(
                second_token, tokenizer)

            if scoreboard_player.player_type == PlayerType.INTEGER:
                if not isinstance(scoreboard_player.value, int):
                    raise ValueError("scoreboard_player.value is not int")
                compared = f'score {first_token.string} {DataPack.var_name} matches'
                if operator in {'===', '==', '='}:
                    return Condition(
                        f'{compared} {scoreboard_player.value}', IF)
                if operator == '>=':
                    return Condition(
                        f'{compared} {scoreboard_player.value}..', IF)
                if operator == '>':
                    return Condition(
                        f'{compared} {scoreboard_player.value+1}..', IF)
                if operator == '<=':
                    return Condition(
                        f'{compared} ..{scoreboard_player.value}', IF)
                if operator == '<':
                    return Condition(
                        f'{compared} ..{scoreboard_player.value-1}', IF)
            else:
                if operator == '!=':
                    if isinstance(scoreboard_player.value, int):
                        raise ValueError("scoreboard_player.value is int")
                    return Condition(
                        f'score {first_token.string} {DataPack.var_name} = {scoreboard_player.value[1]} {scoreboard_player.value[0]}', UNLESS)

                if operator in {'===', '==', '='}:
                    operator = '='

                if isinstance(scoreboard_player.value, int):
                    raise ValueError("scoreboard_player.value is int")
                return Condition(
                    f'score {first_token.string} {DataPack.var_name} {operator} {scoreboard_player.value[1]} {scoreboard_player.value[0]}', IF)
            break

        if tokens[1].token_type == TokenType.KEYWORD and tokens[1].string == 'matches':
            if len(tokens) > 3:
                raise JMCSyntaxException(
                    "Unexpected token in condition", tokens[3], tokenizer)
            if tokens[2].token_type != TokenType.KEYWORD:
                raise JMCSyntaxException(
                    "Expected keyword", tokens[2], tokenizer)

            match_tokens_ = tokenizer.split_keyword_token(tokens[2], '..')
            match_tokens = tokenizer.find_token(match_tokens_, '..')
            if len(match_tokens) != 2 or len(
                    match_tokens[0]) > 1 or len(match_tokens[1]) > 1:
                raise JMCSyntaxException(
                    "Expected <integer>..<integer> after 'matches'", tokens[2], tokenizer)
            if not match_tokens[0]:
                raise JMCSyntaxException(
                    "Expected <integer>..<integer> after 'matches'", tokens[2], tokenizer, suggestion=f"Use {first_token.string}<={match_tokens[1][0].string} instead")
            if not match_tokens[1]:
                raise JMCSyntaxException(
                    "Expected <integer>..<integer> after 'matches'", tokens[2], tokenizer, suggestion=f"Use {first_token.string}>={match_tokens[0][0].string} instead")

            first_int = int(match_tokens[0][0].string)
            second_int = int(match_tokens[1][0].string)
            if first_int == second_int:
                raise JMCSyntaxException(
                    "First integer must not equal second integer after 'matches'", tokens[2], tokenizer, suggestion=f"Use {first_token.string}=={match_tokens[0][0].string} instead")
            if first_int > second_int:
                raise JMCSyntaxException(
                    "First integer must be less than second integer after 'matches'", tokens[2], tokenizer, suggestion=f"Did you mean {match_tokens[1][0].string}..{match_tokens[0][0].string} ?")

            return Condition(
                f'score {first_token.string} {DataPack.var_name} matches {tokens[2].string}', IF)

        raise JMCSyntaxException(
            "Operator not found in custom condition", tokens[0], tokenizer)

    matched_function = BOOL_FUNCTIONS.get(
        tokens[0].string, None)
    if matched_function is not None:
        if len(tokens) > 2:
            raise JMCSyntaxException(
                "Unexpected token", tokens[2], tokenizer, display_col_length=False)

        return Condition(
            *matched_function(tokens[1], datapack, tokenizer).call_bool())
    # End
    conditions: list[str] = []
    if tokens[0].string not in {"block", "blocks",
                                "data", "entity", "predicate", "score"}:
        raise JMCValueError(
            f"Unrecoginized condition '{tokens[0].string}'",
            tokens[0],
            tokenizer,
            suggestion="Consider using 'block' or 'blocks' or 'data' or 'entity' or 'predicate' or 'score'.")
    for token in tokens:
        if token.token_type == TokenType.PAREN_SQUARE:
            if not conditions:
                raise JMCSyntaxException(
                    "Unexpected square parenthesis []", token, tokenizer)
            conditions[-1] += token.string
        else:
            conditions.append(token.string)
    return Condition(' '.join(conditions), IF)


def find_operator(_tokens: list[Token], operator: str,
                  tokenizer: Tokenizer) -> list[list[Token]]:
    """
    Find sepecific operator in tokens and split them

    :param _tokens: List of tokens search
    :param operator: Operator to search for
    :param tokenizer: Tokenizer
    :return: List of (list of tokens)
    """
    list_of_tokens: list[list[Token]] = []
    tokens: list[Token] = []
    if _tokens[0].token_type == TokenType.KEYWORD and _tokens[0].string == operator:
        raise JMCSyntaxException(
            f"Unexpected operator ({operator})", _tokens[0], tokenizer)

    elif _tokens[-1].token_type == TokenType.KEYWORD and _tokens[-1].string == operator:
        raise JMCSyntaxException(
            f"Unexpected operator ({operator})", _tokens[-1], tokenizer)

    for token in _tokens:
        if token.token_type == TokenType.KEYWORD and token.string == operator:
            list_of_tokens.append(tokens)
            tokens = []
        else:
            tokens.append(token)
    list_of_tokens.append(tokens)
    return list_of_tokens


def condition_to_ast(
        tokens: list[Token], tokenizer: Tokenizer, datapack: DataPack) -> AST_TYPE:
    """
    Turn condition in form of list of tokens to abstract syntax tree

    :param tokens: Condition in form of tokens
    :param tokenizer: Tokenizer
    :param datapack: Datapack object
    :raises JMCSyntaxException: Empty round parenthesis inside condition
    :return: Abstract syntax tree
    """
    if len(tokens) == 1 and tokens[0].token_type == TokenType.PAREN_ROUND:
        if tokens[0].string == '()':
            raise JMCSyntaxException(
                "Empty round parenthesis () inside condition", tokens[0], tokenizer)

        tokenizer = Tokenizer(tokens[0].string[1:-1], tokenizer.file_path,
                              tokens[0].line, tokens[0].col + 1, tokenizer.file_string, expect_semicolon=False)
        tokens = tokenizer.programs[0]
    tokens = tokenizer.split_keyword_tokens(tokens, [OR_OPERATOR])
    list_of_tokens = find_operator(tokens, OR_OPERATOR, tokenizer)
    if len(list_of_tokens) > 1:
        return {"operator": OR_OPERATOR, "body": [
            condition_to_ast(tokens, tokenizer, datapack) for tokens in list_of_tokens]}

    tokens = tokenizer.split_keyword_tokens(list_of_tokens[0], [AND_OPERATOR])
    list_of_tokens = find_operator(tokens, AND_OPERATOR, tokenizer)
    if len(list_of_tokens) > 1:
        return {"operator": AND_OPERATOR, "body": [
            condition_to_ast(tokens, tokenizer, datapack) for tokens in list_of_tokens]}

    # NotOperator should have a body as either dict or string and not list
    tokens = tokenizer.split_keyword_tokens(list_of_tokens[0], [NOT_OPERATOR])
    if tokens[0].token_type == TokenType.KEYWORD and tokens[0].string == NOT_OPERATOR:
        return {"operator": NOT_OPERATOR, "body":
                condition_to_ast(tokens[1:], tokenizer, datapack)}

    return custom_condition(tokens, tokenizer, datapack)


def ast_to_commands(
        ast: AST_TYPE, datapack: DataPack) -> tuple[list[Condition], list[tuple[list[Condition], int]] | None]:
    """
    Parse abstract syntax tree into list of conditions and list of commands that need to come before for it to works

    :param ast: Abstract syntax tree
    :raises ValueError: Invalid AST
    :return: A tuple of (
        A chain of conditions(List of Condition)

        Commands(
            List of Condition and n (`__logic__n` for minecraft function name)
        ) that need to come before (can be None)
    )
    """
    if isinstance(ast, Condition):
        return [ast], None

    if ast["operator"] == AND_OPERATOR:
        conditions: list[Condition] = []
        precommand_and: list[tuple[list[Condition], int]] = []
        if isinstance(ast["body"], Condition):
            raise ValueError(
                'ast["body"] is a Condition instead of list in AND')
        for and_body in ast["body"]:
            if isinstance(and_body, str):
                raise ValueError('ast["body"] is string')
            _conditions, precommand = ast_to_commands(and_body, datapack)  # noqa
            conditions.extend(_conditions)
            if precommand is not None:
                precommand_and.extend(precommand)
        return conditions, (precommand_and if precommand_and else None)

    elif ast["operator"] == OR_OPERATOR:
        _count = datapack.data.condition_count
        datapack.data.condition_count += 1
        precommand_or: list[tuple[list[Condition], int]] = []
        if isinstance(ast["body"], Condition):
            raise ValueError(
                'ast["body"] is a Condition instead of list in OR')
        for or_body in ast["body"]:
            if isinstance(or_body, str):
                raise ValueError('ast["body"] is string')
            conditions, precommand = ast_to_commands(or_body, datapack)
            if precommand is not None:
                precommand_or.extend(precommand)
            precommand_or.append((conditions, _count))

        return [Condition(
            f"score {VAR}{_count} {DataPack.var_name} matches 1", IF)], precommand_or

    elif ast["operator"] == NOT_OPERATOR:
        if isinstance(ast["body"], Condition):
            conditions, precommand = ast_to_commands(ast["body"], datapack)
        elif isinstance(ast["body"], dict):
            if ast["body"]["operator"] == OR_OPERATOR:
                ast["body"]["operator"] = AND_OPERATOR
            elif ast["body"]["operator"] == AND_OPERATOR:
                ast["body"]["operator"] = OR_OPERATOR
            conditions, precommand = ast_to_commands(ast["body"], datapack)
        else:
            raise ValueError('ast["body"] is a list or str')
        for condition in conditions:
            condition.reverse()
        return conditions, precommand

    raise ValueError("Invalid AST")


def ast_to_strings(ast: AST_TYPE, datapack: DataPack) -> tuple[str, str]:
    """
    Turns AST into tuple of full `execute if command` a multiple line string representing precommands

    :param ast: Abstract Syntax tree
    :return: tuple of `execute if` command(excluding `execute`) a multiple line string representing precommands
    """

    conditions, precommand_conditions = ast_to_commands(ast, datapack)
    if precommand_conditions is None:
        precommand = ""
    else:
        precommands = []
        current_count = -1
        for conditions_and_count in precommand_conditions:
            if conditions_and_count[1] > current_count:
                current_count += 1
                precommands.append(
                    f"scoreboard players set {VAR}{current_count} {DataPack.var_name} 0")
                precommands.append(
                    f"execute {merge_condition(conditions_and_count[0])} run scoreboard players set {VAR}{current_count} {DataPack.var_name} 1")
                continue

            precommands.append(
                f"execute unless score {VAR}{current_count} {DataPack.var_name} matches 1 {merge_condition(conditions_and_count[0])} run scoreboard players set {VAR}{current_count} {DataPack.var_name} 1")
        precommand = '\n'.join(precommands)

    condition_string = merge_condition(conditions)
    return condition_string, precommand


def parse_condition(condition_token: Token |
                    list[Token], tokenizer: Tokenizer, datapack: DataPack) -> tuple[str, str]:
    """
    Parse condition token(s) (token or list of tokens) to `if ...` and pre-commands with newline
    Example:
    ```py
    condition1, precommands1 = parse_condition(...)
    condition2, precommands2 = parse_condition(...)
    commands = [
        f"{precommands1}execute {condition1} run {datapack.add_private_function("if_else", token, tokenizer)}",
        f"{precommands2}execute unless ... {condition2} run {datapack.add_private_function("if_else", token, tokenizer)}",
    ]
    return datapack.add_raw_private_function("if_else", commands)
    ```

    :param condition_token: Token or List of tokens
    :param tokenizer: Tokenizer
    :param datapack: Datapack object
    :return: tuple of `execute if` command(excluding `execute`) a multiple line string representing precommands
    """
    datapack.data.condition_count = 0
    tokens = condition_token if isinstance(
        condition_token, list) else [condition_token]

    ast = condition_to_ast(tokens, tokenizer, datapack)
    condition, precommand = ast_to_strings(ast, datapack)
    precommand = precommand + '\n' if precommand else ""
    return condition, precommand



"""Utility for compiling"""
import functools
import re
from json import JSONEncoder, dumps
from typing import TYPE_CHECKING, Any

from .exception import JMCSyntaxException, MinecraftSyntaxWarning

if TYPE_CHECKING:
    from .tokenizer import Token, Tokenizer


class SingleTonMeta(type):
    """
    Metaclass for singleton
    """
    _instances: dict["SingleTonMeta", Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class SingleTon(metaclass=SingleTonMeta):
    """
    Super class for a singleton class

    :raises TypeError: Instantiated directly
    """

    def __init__(self) -> None:
        pass

    def __new__(cls, *args, **kwargs):
        if cls is SingleTon:
            raise TypeError(
                f"Only children of '{cls.__name__}' may be instantiated")
        return super().__new__(cls, *args, **kwargs)


def is_number(string: str) -> bool:
    """Whether string can be converted to integer"""
    try:
        int(string)
    except ValueError:
        return False
    else:
        return True


def is_float(string: str) -> bool:
    """Whether string can be converted to integer"""
    try:
        float(string)
    except ValueError:
        return False
    else:
        return True


def is_connected(current_token: "Token", previous_token: "Token") -> bool:
    """Whether 2 tokens are next to each other"""
    return (
        previous_token.line == current_token.line and
        previous_token.col +
        previous_token.length == current_token.col
    )


def __parse_to_string(
        token: "Token", tokenizer: "Tokenizer") -> dict[str, str | bool | dict[str, str]]:
    """
    Parse `toString` in JMC

    :param token: paren_round token (arguments of `toString`)
    :param tokenizer: token's tokenizer
    :return: Dictionary of key(key) and value(string or boolean)
    """
    json: dict[str, str | bool | dict[str, str]] = {}
    if token.string == '()':
        return json

    for kwargs in token.string[1:-1].split(','):
        if '=' not in kwargs:
            raise JMCSyntaxException(
                "'=' not found in .toString argument", token, tokenizer)
        key, value, *extras = kwargs.split('=')
        key = key.strip()
        if extras:
            raise JMCSyntaxException(
                f"Too many '=' found in .toString argument (got {len(extras)+1})", token, tokenizer)
        if value in {'""', "''"}:
            raise JMCSyntaxException(
                ".toString function accepts keyword as arguments, not string", token, tokenizer, suggestion="Do not use empty string as argument")
        if value[0] in {'"', "'"}:
            raise JMCSyntaxException(
                ".toString function accepts keyword as arguments, not string", token, tokenizer, suggestion=f"Use '{value[1:-1]}' instead of '{value}'")
        if value[0] in {'{', '(', '['}:
            raise JMCSyntaxException(
                "Brackets are not supported in .toString", token, tokenizer, suggestion="clickEvent and hoverEvent are not supported, use vanila JSON instead.")

        if key in {'font', 'color'}:
            json[key] = value
        elif key in {'bold', 'italic', 'underlined', 'strikethrough', 'obfuscated'}:
            if value not in {'true', 'false'}:
                raise JMCSyntaxException(
                    f"value of {key} .toString must be either 'true' or 'false'", token, tokenizer)
            json[key] = value == 'true'
        elif key in {'clickEvent', 'hoverEvent'}:
            raise JMCSyntaxException(
                "clickEvent and hoverEvent are not supported in .toString", token, tokenizer, suggestion="Use vanila JSON instead")
        elif key == 'text':
            raise JMCSyntaxException(
                "'text' key is incompatible with 'score' in .toString", token, tokenizer)
        else:
            raise JMCSyntaxException(
                f"Unrecognized key in .toString (got {key})", token, tokenizer, suggestion="Avaliable keys are 'font', 'color', 'bold', 'italic', 'underlined', 'strikethrough', 'obfuscated'")
    return json


def __search_to_string(match: re.Match, token: "Token",
                       var_name: str, tokenizer: "Tokenizer") -> str:
    """
    Function for regex.subn

    :param match: Match object
    :param token: Token to search for `toString`
    :param var_name: `DataPack.VARNAME`
    :param tokenizer: token's tokenizer
    :return: JSON formatted string
    """
    var: str = match.group(1)
    properties = __parse_to_string(token, tokenizer)
    properties["score"] = {"name": var, "objective": var_name}
    return dumps(properties)


def search_to_string(last_str: str, token: "Token",
                     var_name: str, tokenizer: "Tokenizer") -> tuple[str, bool]:
    """
    Find `toString` in a last_str

    :param last_str: Last string before this token
    :param token: paren_round token (Arguments of `toString`)
    :param VAR_NAME: `DataPack.VARNAME``
    :param tokenizer: token's tokenizer
    :return: Tuple of New string and Whether `toString` is found
    """
    new_str, count = re.subn(
        r'(\$[A-z0-9\-\.\_]+)\.toString$', lambda match: __search_to_string(match, token, var_name, tokenizer), last_str)
    if count:
        return new_str, True
    return last_str, False


class JSONUniversalEncoder(JSONEncoder):
    """
    JSONEncoder that can encode everything, used for displaying results
    """

    def default(self, o):
        try:
            iterator = iter(o)
        except TypeError:
            pass
        else:
            return list(iterator)

        return repr(o)


def monitor_results(func):
    """
    Decorator to monitor a function for debugging

    :param func: Function for decorator
    :return: Wrapper function
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return_value = func(*args, **kwargs)
        print(f"""Function call {func.__name__}(
args={dumps(args, indent=2, cls=JSONUniversalEncoder)},
kwargs={dumps(kwargs, indent=2, cls=JSONUniversalEncoder)}
) returns {dumps(return_value, indent=2, cls=JSONUniversalEncoder)}
""")
        return return_value
    return wrapper


def convention_jmc_to_mc(token: "Token", tokenizer: "Tokenizer",
                         is_make_lower: bool = True, substr: tuple[int, int] | None = None) -> str:
    """
    Turns JMC function/predicate name syntax to vanilla's syntax

    .. example::
    >>> convention_jmc_to_mc(Token.empty("hello.World"))
    "hello_world"

    :param token: keyword token
    :param tokenizer: token's tokenizer
    """
    string = token.string
    if substr is not None:
        string = string[substr[0]:substr[1]]
    if string.startswith("."):
        raise JMCSyntaxException("Name started with '.'", token, tokenizer)
    if string.endswith("."):
        raise JMCSyntaxException("Name ended with '.'", token, tokenizer)
    if is_make_lower:
        string = string.lower()
    if re.match('^[a-z0-9_\\.]+$', string) is None:
        raise MinecraftSyntaxWarning(
            f"Invalid character detected in '{string}'", token, tokenizer
        )
    return string.replace('.', '/')



from enum import Enum
from getpass import getpass
from threading import Event
from ..compile import Logger

logger = Logger(__name__)


class Colors(Enum):
    HEADER = '\033[1;33;40m'
    YELLOW = '\033[33;40m'
    INFO = '\033[94;40m'
    INPUT = '\033[96;40m'
    PURPLE = '\033[35;40m'
    FAIL = '\033[91;40m'
    FAIL_BOLD = '\033[1;91;40m'
    ENDC = '\033[0;0;40m'
    EXIT = '\033[0;0;0m'
    NONE = '\033[0;37;40m'


def pprint(values, color: Colors = Colors.NONE):
    """
    Print but with colors

    :param values: Value for printing
    :param color: color of printing
    """
    print(f"{color.value}{values}{Colors.ENDC.value}")


def get_input(prompt: str = "> ", color: Colors = Colors.INPUT) -> str:
    """
    Get an input from user

    :param promt: Display string infront
    :param color: Color of the input and promt
    :return: input from user
    """
    input_value = input(f"{color.value}{prompt}")
    print(Colors.ENDC.value, end="")
    logger.info(f"Input from user: {input_value}")
    return input_value


def press_enter(prompt: str, color: Colors = Colors.INPUT) -> None:
    """
    Wait for Enter key from user

    :param prompt: Display string
    :param color: Color of prompt
    """
    getpass(f"{color.value}{prompt}{Colors.ENDC.value}")


def error_report(error: Exception) -> None:
    """
    Report error to user

    :param error: Exception for reporting
    """
    pprint(type(error).__name__, Colors.FAIL_BOLD)
    pprint(error, Colors.FAIL)


def handle_exception(error: Exception, event: Event, is_ok: bool):
    """
    Tell user when unexpected crash happens and reset

    :param error: Exception
    :param event: Event object to stop compiling
    :param is_ok: Whether to tell user this error is normal
    """
    event.set()
    pprint("Unexpected error causes program to crash", Colors.FAIL)
    pprint(type(error).__name__, Colors.FAIL_BOLD)
    pprint(error, Colors.FAIL)
    if not is_ok:
        pprint(
            "NOTE: This shouldn't happen. Please contact WingedSeal.",
            Colors.FAIL)
    logger.critical("Program crashed")
    logger.exception("")
    press_enter("Press Enter to continue...")


class RestartException(BaseException):
    """Raise to restart the program without telling error to user"""



"""Module containing JMCFunction subclasses for custom JMC function that returns a minecraft integer to a scoreboard variable"""

from ...datapack import DataPack
from ..utils import ArgType
from ..jmc_function import JMCFunction, FuncType, func_property
from ...exception import JMCValueError


@func_property(
    func_type=FuncType.VARIABLE_OPERATION,
    call_string='Math.sqrt',
    arg_type={
        "n": ArgType.SCOREBOARD
    },
    name='math_sqrt'
)
class MathSqrt(JMCFunction):
    def call(self) -> str:
        x = '__math__.x'
        x_n = '__math__.x_n'
        x_n_sq = '__main__.x_n_sq'
        N = '__math__.N'
        diff = '__math__.different'
        var = DataPack.var_name
        if self.is_never_used():
            self.datapack.add_int(2)
            self.datapack.add_raw_private_function(
                self.name,
                [
                    f"scoreboard players operation {x} {var} = {x_n} {var}",
                    f"scoreboard players operation {x_n} {var} = {N} {var}",
                    f"scoreboard players operation {x_n} {var} /= {x} {var}",
                    f"scoreboard players operation {x_n} {var} += {x} {var}",
                    f"scoreboard players operation {x_n} {var} /= 2 {DataPack.int_name}",
                    f"scoreboard players operation {diff} {var} = {x} {var}",
                    f"scoreboard players operation {diff} {var} -= {x_n} {var}",
                    f"execute unless score {diff} {var} matches 0..1 run {self.datapack.call_func(self.name, 'newton_raphson')}",
                ],
                'newton_raphson'
            )
            self.datapack.add_raw_private_function(
                self.name,
                [
                    f"scoreboard players set {x_n} {var} 1225",
                    self.datapack.call_func(self.name, 'newton_raphson'),
                    f"scoreboard players operation {x_n_sq} {var} = {x_n} {var}",
                    f"scoreboard players operation {x_n_sq} {var} *= {x_n} {var}",
                    f"scoreboard players operation {x_n} {var} /= 2 {DataPack.int_name}",
                    f"scoreboard players operation {diff} {var} = {x} {var}",
                    f"scoreboard players operation {diff} {var} -= {x_n} {var}",
                    f"execute if score {x_n_sq} {var} > {N} {var} run scoreboard players remove {x_n} {var} 1",
                ],
                'main'
            )

        run = [
            f"scoreboard players operation {N} {var} = {self.args['n']}",
            self.datapack.call_func(self.name, 'main'),
            f"scoreboard players operation {self.var} {var} = {x_n} {var}"
        ]

        if self.is_execute:
            count = self.datapack.get_count(self.name)
            return self.datapack.add_raw_private_function(
                self.name,
                run,
                count
            )
        return '\n'.join(run)


@func_property(
    func_type=FuncType.VARIABLE_OPERATION,
    call_string='Math.random',
    arg_type={
        "min": ArgType.SCOREBOARD_PLAYER,
        "max": ArgType.SCOREBOARD_PLAYER
    },
    name='math_random',
    defaults={
        "min": "1",
        "max": "2147483647"
    }
)
class MathRandom(JMCFunction):
    def call(self) -> str:
        seed = '__math__.seed'
        a = '__math__.rng.a'
        c = '__math__.rng.c'
        var = DataPack.var_name
        start = int(self.args["min"])
        end = int(self.args["max"])
        if end < start:
            raise JMCValueError(
                f"max cannot be less than min in {self.call_string}", self.token, self.tokenizer, suggestion="Try swapping max and min")
        if self.is_never_used():
            self.datapack.add_load_command(
                f"""execute unless score {seed} {var} matches -2147483648..2147483647 run {
                    self.datapack.add_raw_private_function(
                        self.name,
                        [
                            f'summon area_effect_cloud ~ ~ ~ {{Tags:["{self.datapack.private_name}.{self.name}"]}}',
                            f"execute store result score {seed} {var} run data get entity @e[limit=1,type=area_effect_cloud,tag={self.datapack.private_name}.{self.name}] UUID[0] 1",
                            f"kill @e[type=area_effect_cloud,tag={self.datapack.private_name}.{self.name}]",
                            f"scoreboard players set {a} {var} 656891",
                            f"scoreboard players set {c} {var} 875773"
                        ],
                        'setup'
                    )
                }""")
            self.datapack.add_raw_private_function(
                self.name,
                [
                    f"execute if score {seed} {var} matches ..0 run scoreboard players add {seed} {var} 2147483647",
                    f"scoreboard players operation {seed} {var} *= {a} {var}",
                    f"scoreboard players operation {seed} {var} += {c} {var}"
                ],
                'main'
            )

        mod = end - start + 1
        self.datapack.add_int(mod)

        run = [
            self.datapack.call_func(self.name, 'main'),
            f"scoreboard players operation {self.var} {var} = {seed} {var}",
            f"scoreboard players operation {self.var} {var} %= {mod} {DataPack.int_name}"
        ]

        if start:
            run.append(f"scoreboard players add {self.var} {var} {start}")

        if self.is_execute:
            count = self.datapack.get_count(self.name)
            return self.datapack.add_raw_private_function(
                self.name,
                run,
                count
            )
        return '\n'.join(run)



"""Module responsibile for all compiling in jmc"""
from json import JSONDecodeError, dump, dumps, loads
from shutil import rmtree
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .header import parse_header, Header
from .lexer import Lexer
from .log import Logger
from .datapack import DataPack
from .exception import JMCBuildError

if TYPE_CHECKING:
    from ..terminal import Configuration

logger = Logger(__name__)
JMC_CERT_FILE_NAME = 'jmc.txt'


def compile_jmc(config: "Configuration", debug: bool = False) -> None:
    """
    Compile the files and build the datapack

    :param config: Configuration dictionary
    :param debug: Whether to debug into log, defaults to False
    """
    logger.info("Configuration:\n" + dumps(config.toJSON(), indent=2))
    Header.clear()
    read_cert(config)
    read_header(config)
    logger.info("Parsing")
    lexer = Lexer(config)
    if debug:
        logger.info(f'Datapack :{lexer.datapack!r}')
    build(lexer.datapack, config)


def cert_config_to_string(cert_config: dict[str, str]) -> str:
    """
    Turns certificate configuration dictionary into a string for output

    :param cert_config: Certificate configuration dictionary
    :return: Converted string
    """
    return '\n'.join([f"{key}={value}" for key, value in cert_config.items()])


def string_to_cert_config(string: str) -> dict[str, str]:
    """
    Turns string into certificate configuration dictionary for further read

    :param string: String for convertion
    :return: Converted cert_config
    """
    cert_config = {}
    for line in string.split('\n'):
        if not line or line.isspace():
            continue
        key, value = line.split('=')
        cert_config[key.strip()] = value.strip()
    return cert_config


def make_cert(cert_config: dict[str, str], path: Path) -> None:
    """
    Write certificate file

    :param cert_config: Certificate configuration
    :param path: Path to write `cert_config` to
    """
    path.parent.mkdir(parents=True, exist_ok=False)
    with path.open('w+') as file:
        file.write(cert_config_to_string(cert_config))


def get_cert() -> dict[str, str]:
    """
    Make a new Certificate configuration from current DataPack class info

    :return: Certificate configuration
    """
    return {
        "LOAD": DataPack.load_name,
        "TICK": DataPack.tick_name,
        "PRIVATE": DataPack.private_name,
        "VAR": DataPack.var_name,
        "INT": DataPack.int_name
    }


def read_header(config: "Configuration",
                _test_file: str | None = None) -> bool:
    """
    Read the main header file

    :param config: JMC configuration
    :return: Whether the main header file was found
    """
    header = Header()
    header_file = Path(config.target_str[:-len(".jmc")] + ".hjmc")
    parent_target = Path(config.target_str).parent
    if header_file.is_file() or _test_file is not None:
        header.add_file_read(header_file)
        logger.info("Header file found.")
        if _test_file is None:
            with header_file.open('r', encoding="utf-8") as file:
                header_str = file.read()
        else:
            header_str = _test_file
        logger.info(f"Parsing {header_file}")
        parse_header(header_str, header_file.as_posix(), parent_target)
        return True

    logger.info("Header file not found.")
    return False


def read_cert(config: "Configuration", _test_file: str | None = None):
    """
    Read Certificate(JMC.txt)

    :param config: JMC configuration
    :raises JMCBuildError: Can't find JMC.txt
    """
    namespace_folder = Path(config.output) / 'data' / config.namespace
    cert_file = namespace_folder / JMC_CERT_FILE_NAME
    old_cert_config = get_cert()
    if namespace_folder.is_dir() or _test_file is not None:
        if not cert_file.is_file() and _test_file is None:
            raise JMCBuildError(
                f"{JMC_CERT_FILE_NAME} file not found in namespace folder.\n To prevent accidental overriding of your datapack please delete the namespace folder yourself.")
        if _test_file is None:
            with cert_file.open('r') as file:
                cert_str = file.read()
        else:
            cert_str = _test_file
        try:
            cert_config = string_to_cert_config(cert_str)
        except ValueError:
            cert_config = {}
        DataPack.load_name = cert_config.get(
            "LOAD", old_cert_config["LOAD"])
        DataPack.tick_name = cert_config.get(
            "TICK", old_cert_config["TICK"])
        DataPack.private_name = cert_config.get(
            "PRIVATE", old_cert_config["PRIVATE"])
        DataPack.var_name = cert_config.get(
            "VAR", old_cert_config["VAR"])
        DataPack.int_name = cert_config.get(
            "INT", old_cert_config["INT"])
        cert_config = get_cert()
        if _test_file is None:
            rmtree(namespace_folder.resolve().as_posix())
    else:
        cert_config = old_cert_config
    if _test_file is None:
        make_cert(cert_config, cert_file)


def read_func_tag(path: Path, config: "Configuration") -> dict[str, Any]:
    """
    Read minecraft function tag file

    :param path: Path to minecraft function tag file
    :param config: JMC configuration
    :raises JMCBuildError: MalformedJsonException
    :raises JMCBuildError: Can't find `values` key in json
    :return: Content of function tag file in dictionary
    """
    if path.is_file():
        with path.open('r') as file:
            content = file.read()
        try:
            json: dict[str, Any] = loads(content)
            json["values"] = [
                value for value in json["values"] if not value.startswith(config.namespace + ':')]
        except JSONDecodeError as error:
            raise JMCBuildError(
                f"MalformedJsonException: Cannot parse {path.resolve().as_posix()}. Deleting the file to reset.") from error
        except KeyError as error:
            raise JMCBuildError(
                f'"values" key not found in {path.resolve().as_posix()}. Deleting the file to reset.') from error
    else:
        json = {"values": []}
    return json


def post_process(string: str) -> str:
    """
    Post processing of .mcfunction files

    - Add credits at the end of every .mcfunction file

    :param string: File content
    :return: Processed file content
    """
    header = Header()
    if not header.credits:
        return string

    string += "\n" * 2
    for line in header.credits:
        string += f"\n# {line}" if line else "\n#"
    return string


def build(datapack: DataPack, config: "Configuration",
          _is_virtual: bool = False) -> dict[str, str] | None:
    """
    Build and write files for minecraft datapack

    :param datapack: DataPack object
    :param config: JMC configuration
    :param _is_virtual: Whether to make a dictionary of output result instead of writing to files
    :returns: Dictionary of file path and file content if _is_virtual is True
    """
    output: dict[str, Any] = {}

    logger.debug(f"Building (_is_virtual={_is_virtual})")
    datapack.build()
    output_folder = Path(config.output)
    namespace_folder = output_folder / 'data' / config.namespace
    functions_tags_folder = output_folder / \
        'data' / 'minecraft' / 'tags' / 'functions'

    if not _is_virtual:
        functions_tags_folder.mkdir(exist_ok=True, parents=True)
    load_tag = functions_tags_folder / 'load.json'
    tick_tag = functions_tags_folder / 'tick.json'

    load_json = {"values": []} if _is_virtual else read_func_tag(
        load_tag, config)
    tick_json = {"values": []} if _is_virtual else read_func_tag(
        tick_tag, config)

    load_json["values"].append(f'{config.namespace}:{DataPack.load_name}')
    if _is_virtual:
        output[load_tag.as_posix()] = dumps(load_json, indent=2)
    else:
        with load_tag.open('w+') as file:
            dump(load_json, file, indent=2)

    if DataPack.tick_name in datapack.functions and datapack.functions[DataPack.tick_name]:
        tick_json["values"].append(
            f'{config.namespace}:{DataPack.tick_name}')
        if _is_virtual:
            output[tick_tag.as_posix()] = dumps(tick_json, indent=2)
        else:
            with tick_tag.open('w+') as file:
                dump(tick_json, file, indent=2)

    for func_path, func in datapack.functions.items():
        path = namespace_folder / 'functions' / (func_path + '.mcfunction')
        content = post_process(func.content)
        if content:
            if _is_virtual:
                output[path.as_posix()] = content
            else:
                path.parent.mkdir(parents=True, exist_ok=True)
                with path.open('w+') as file:
                    file.write(content)

    for json_path, json in datapack.jsons.items():
        # if json_path.startswith("minecraft/"):
        #     path = namespace_folder / (json_path + '.json')
        # else:
        #     # len("minecraft/") = 10
        #     path = output_folder / 'data' / 'minecraft' / (json_path[10:] + '.json')
        path = namespace_folder / (json_path + '.json')
        if json:
            if _is_virtual:
                output[path.as_posix()] = dumps(json, indent=2)
            else:
                path.parent.mkdir(parents=True, exist_ok=True)
                with path.open('w+') as file:
                    dump(json, file, indent=2)
    if _is_virtual:
        return output

    with (output_folder / 'pack.mcmeta').open('w+') as file:
        dump({
            "pack": {
                "pack_format": int(config.pack_format),
                "description": config.description
            }
        }, file, indent=2)
    return None



"""Module responsible for handling all Function Content parsing in Lexer"""
from typing import TYPE_CHECKING
from json import dumps

from .vanilla_command import COMMANDS as VANILLA_COMMANDS
from .tokenizer import Tokenizer, Token, TokenType
from .exception import JMCSyntaxException, MinecraftSyntaxWarning
from .log import Logger
from .utils import convention_jmc_to_mc, is_number, is_connected, search_to_string
from .datapack import DataPack
from .command.condition import BOOL_FUNCTIONS
from .command import (FLOW_CONTROL_COMMANDS,
                      variable_operation,
                      JMCFunction,
                      FuncType,
                      )

if TYPE_CHECKING:
    from .lexer import Lexer

logger = Logger(__name__)

EXECUTE_EXCLUDED_COMMANDS = JMCFunction.get_subclasses(
    FuncType.EXECUTE_EXCLUDED)
"""Dictionary of command's name and a class of JMCFunction type for custom jmc command that can't be used with `/execute`"""
LOAD_ONCE_COMMANDS = JMCFunction.get_subclasses(FuncType.LOAD_ONCE)
"""Dictionary of command's name and a class of JMCFunction type for custom jmc command that can be only used *once* in load"""
JMC_COMMANDS = JMCFunction.get_subclasses(FuncType.JMC_COMMAND)
"""Dictionary of command's name and a class of JMCFunction type for custom jmc command"""
LOAD_ONLY_COMMANDS = JMCFunction.get_subclasses(FuncType.LOAD_ONLY)
"""Dictionary of command's name and a class of JMCFunction type for custom jmc command that can only be used in load"""

FIRST_ARGUMENTS = {
    *FLOW_CONTROL_COMMANDS,
    *LOAD_ONCE_COMMANDS,
    *LOAD_ONLY_COMMANDS,
    *JMC_COMMANDS,
    *FLOW_CONTROL_COMMANDS,
    *EXECUTE_EXCLUDED_COMMANDS,
    *VANILLA_COMMANDS
}
"""Set of all vanilla commands and JMC custom syntax
- `if` is excluded since it can be used in execute
- `give` is exluced since it can also be arguments (`/effect give`)"""

FIRST_ARGUMENTS_EXCEPTION = {
    "give": {"effect", "recipe"},
    "clear": {"effect"},
    "if": {"execute"},
    "function": {"schedule"},
    "trigger": {"scoreboard"}
}
"""Dictionary of (FIRST_ARGUMENTS that can also be used as normal argument in a command) and (those commands)"""


ALLOW_NUMBER_AFTER_CURLY_PAREN = {
    'give',
    'clear'
}
"""List of vanilla command to stop JMC from terminating line from curly parenthesis (Allow number after curly parenthesis)"""


def append_commands(commands: list[str], string: str) -> None:
    """
    Append a new argument to a comand(list of minecraft arguments)

    :param commands: Entire command(list of minecraft arguments) to add to
    :param string: A new argument to add
    """
    commands.append(string)


class FuncContent:
    """
    A class representation of a row function for parsing content inside the function

    :param tokenizer: Tokenizer
    :param programs: List of commands(List of arguments(Token))
    :param is_load: Whether the function is a load function
    """
    __slots__ = 'tokenizer', 'programs', 'is_load', 'command_strings', 'commands', 'command', 'is_expect_command', 'is_execute', 'lexer'

    command: list[Token]
    is_expect_command: bool
    is_execute: bool

    def __init__(self, tokenizer: Tokenizer,
                 programs: list[list[Token]], is_load: bool, lexer: "Lexer") -> None:

        self.tokenizer = tokenizer
        self.programs = programs
        self.is_load = is_load
        self.command_strings: list[str] = []
        self.commands: list[str] = []
        self.lexer = lexer

    def parse(self) -> list[str]:
        """
        Parse the content
        :return: List of commands(String)
        """
        for command in self.programs:
            self.command = command
            if self.command[0].token_type != TokenType.KEYWORD:
                raise JMCSyntaxException(
                    "Expected keyword", self.command[0], self.tokenizer)
            if self.command[0].string == 'new':
                raise JMCSyntaxException(
                    "'new' keyword found in function", self.command[0], self.tokenizer)
            if self.command[0].string == 'class':
                raise JMCSyntaxException(
                    "'class' keyword found in function", self.command[0], self.tokenizer)
            if self.command[0].string == 'function' and len(self.command) == 4:
                raise JMCSyntaxException(
                    "Function declaration found in function", self.command[0], self.tokenizer)
            if self.command[0].string == '@import':
                raise JMCSyntaxException(
                    "Importing found in function", self.command[0], self.tokenizer)

            # Boxes check
            if self.lexer.do_while_box is not None:
                if self.command[0].string != 'while':
                    raise JMCSyntaxException(
                        "Expected 'while'", self.command[0], self.tokenizer)
            if self.lexer.if_else_box:
                if self.command[0].string != 'else':
                    append_commands(
                        self.commands, self.lexer.parse_if_else(
                            self.tokenizer))

            if self.commands:
                self.command_strings.append(' '.join(self.commands))
                self.commands = []

            self.__parse_commands()
        # End of Program

        # Boxes check
        if self.lexer.do_while_box is not None:
            raise JMCSyntaxException(
                "Expected 'while'", self.programs[-1][-1], self.tokenizer)
        if self.lexer.if_else_box:
            append_commands(
                self.commands,
                self.lexer.parse_if_else(
                    self.tokenizer))

        if self.commands:
            self.command_strings.append(' '.join(self.commands))
            self.commands = []
        return self.command_strings

    def __parse_commands(self) -> None:
        """Parse command in self.commands"""
        self.is_expect_command = True
        self.is_execute = (self.command[0].string == 'execute')
        command_pos = None

        for key_pos, token in enumerate(self.command):
            if not self.is_expect_command:
                if command_pos is None:
                    raise ValueError("Unknown command_pos")
                self.__not_expect_command(key_pos, token, command_pos)
                continue
            command_pos = key_pos
            if self.__expect_command(key_pos, token):
                break

    def __optimize(self, token: Token) -> bool:
        """
        Optimize minecraft command by removing redundancy
        :param token: Current token
        :return: Whether to cancel appending this token
        """
        if token.token_type != TokenType.KEYWORD:
            return False
        if not self.commands:
            return False

        # `execute as @s`
        if (
            token.string == '@s' and
            self.commands[-1] == 'as' and
            self.commands[-2] not in {'rotated', 'positioned'}
        ):

            self.commands[-1] = 'if entity'
            return False

        # `execute run`
        if (
            token.string == 'run'
            and self.commands[-1] == 'execute'
        ):
            del self.commands[-1]
            return True

        # `run execute`
        if (
            token.string == 'execute'
            and self.commands[-1] == 'run'
            and self.commands[0] == 'execute'
        ):
            del self.commands[-1]
            return True

        return False

    def __not_expect_command(
            self, key_pos: int, token: Token, command_pos: int) -> None:
        """
        Called when expecting an argument of command

        :param key_pos: Current index in self.command
        :param token: token
        :param key_pos: Index of latest command in self.command
        :return: Whether to break out of the loop
        """
        if token.string == 'run' and token.token_type == TokenType.KEYWORD:
            if not self.is_execute:
                raise MinecraftSyntaxWarning(
                    "'run' keyword found outside 'execute' command", token, self.tokenizer)
            self.is_expect_command = True

        if (
            token.token_type == TokenType.KEYWORD and
            token.string in FIRST_ARGUMENTS and
            not (
                token.string in FIRST_ARGUMENTS_EXCEPTION
                and
                self.commands[command_pos] in FIRST_ARGUMENTS_EXCEPTION[token.string]
            )
        ):
            raise JMCSyntaxException(
                f"Keyword({token.string}) at line {token.line} col {token.col} is recognized as a command.\nExpected semicolon(;)", self.command[key_pos - 1], self.tokenizer, col_length=True)

        if self.__optimize(token):
            return

        if token.token_type == TokenType.PAREN_ROUND:
            self.commands[-1], success = search_to_string(
                self.commands[-1], token, DataPack.var_name, self.tokenizer)
            if not success:
                if is_connected(token, self.command[key_pos - 1]):
                    self.commands[-1] += self.lexer.clean_up_paren_token(
                        token, self.tokenizer)
                else:
                    append_commands(
                        self.commands, self.lexer.clean_up_paren_token(token, self.tokenizer))
        elif token.token_type in {TokenType.PAREN_CURLY, TokenType.PAREN_SQUARE}:
            if is_connected(token, self.command[key_pos - 1]):
                self.commands[-1] += self.lexer.clean_up_paren_token(
                    token, self.tokenizer)
            else:
                append_commands(
                    self.commands, self.lexer.clean_up_paren_token(token, self.tokenizer))
        elif token.token_type == TokenType.STRING:
            append_commands(self.commands, dumps(token.string))
        else:
            append_commands(self.commands, token.string)

    def __expect_command(self, key_pos: int, token: Token) -> bool:
        """
        Called when expecting a command

        :param key_pos: Current index in self.command
        :param token: token
        :return: Whether to break out of the loop
        """
        self.is_expect_command = False

        # Handle Errors
        if token.token_type != TokenType.KEYWORD:
            if token.token_type == TokenType.PAREN_CURLY and self.is_execute:
                append_commands(self.commands, self.lexer.datapack.add_arrow_function(
                    'anonymous', token, self.tokenizer))
                return True
            raise JMCSyntaxException(
                "Expected keyword", token, self.tokenizer)

        # End Handle Errors

        if is_number(token.string) and key_pos == 0:
            self.__is_number(key_pos, token)
            return True

        if token.string == 'say':
            self.__is_say(key_pos, token)
            return True

        if token.string.startswith(DataPack.VARIABLE_SIGN):
            if self.__is_startswith_varsign(key_pos, token):
                return True

        if self.__is_jmc_function(key_pos, token):
            if len(self.command[key_pos:]) > 2:
                raise JMCSyntaxException(
                    f"Unexpected token({self.command[key_pos+2].string}) after function call. Expected semicolon(;)", self.command[key_pos + 1], self.tokenizer, col_length=True)
            return True

        if self.__is_flow_control_command(key_pos, token):
            return True

        if token.string in {'new', 'class', '@import'} or (
            token.string == 'function'
            and
            len(self.command) > key_pos + 2
        ):
            if self.is_execute:
                raise JMCSyntaxException(
                    f"This feature({token.string}) can only be used in load function", token, self.tokenizer)
        if len(self.command[key_pos:]) >= 2 and self.command[key_pos +
                                                             1].token_type == TokenType.PAREN_ROUND:
            if len(self.command[key_pos:]) > 2:
                raise JMCSyntaxException(
                    f"Unexpected token({self.command[key_pos+2].string}) after function call. Expected semicolon(;)", self.command[key_pos + 1], self.tokenizer, col_length=True)

            if self.command[key_pos + 1].string != '()':
                raise JMCSyntaxException(
                    f"Custom function({token.string})'s parameter is not supported.\nExpected empty bracket", self.command[key_pos + 1], self.tokenizer)
            append_commands(self.commands,
                            f"function {self.lexer.datapack.namespace}:{convention_jmc_to_mc(token, self.tokenizer)}")
            return True

        if token.string not in VANILLA_COMMANDS:
            raise JMCSyntaxException(
                f"Unrecognized command ({token.string})", token, self.tokenizer)

        if self.__optimize(token):
            return False
        append_commands(self.commands, token.string)
        return False

    def __is_number(self, key_pos: int, token: Token) -> None:
        if len(self.command[key_pos:]) > 1:
            raise JMCSyntaxException(
                "Unexpected token", self.command[key_pos + 1], self.tokenizer, display_col_length=False)

        if not self.command_strings:
            raise JMCSyntaxException(
                "Expected command, got number", token, self.tokenizer, display_col_length=False)

        if not self.command_strings[-1].startswith('execute'):
            for _cmd in ALLOW_NUMBER_AFTER_CURLY_PAREN:
                if self.command_strings[-1].startswith(_cmd):
                    self.command_strings[-1] += ' ' + token.string
                    return
            raise JMCSyntaxException(
                "Unexpected number", token, self.tokenizer, display_col_length=False)

        for _cmd in ALLOW_NUMBER_AFTER_CURLY_PAREN:
            if _cmd in self.command_strings[-1]:
                self.command_strings[-1] += ' ' + token.string
                return

        raise JMCSyntaxException(
            "Unexpected number", token, self.tokenizer, display_col_length=False)

    def __is_say(self, key_pos: int, token: Token) -> None:
        if len(self.command[key_pos:]) == 1:
            raise JMCSyntaxException(
                "Expected string after 'say' command", token, self.tokenizer, col_length=True)

        if self.command[key_pos + 1].token_type != TokenType.STRING:
            raise JMCSyntaxException(
                "Expected string after 'say' command", self.command[key_pos + 1], self.tokenizer, display_col_length=False, suggestion="(In JMC, you are required to wrapped say command's argument in quote.)")

        if len(self.command[key_pos:]) > 2:
            raise JMCSyntaxException(
                "Unexpected token", self.command[key_pos + 2], self.tokenizer, display_col_length=False)

        if '\n' in self.command[key_pos + 1].string:
            raise JMCSyntaxException(
                "Newline found in say command", self.command[key_pos + 1], self.tokenizer, suggestion=r"Use '\\n' instead of '\n'")
        append_commands(
            self.commands, f"say {self.command[key_pos+1].string}")

    def __is_startswith_varsign(self, key_pos: int, token: Token) -> bool:
        if len(
                self.command) > key_pos + 1 and self.command[key_pos + 1].string == 'run' and self.command[key_pos + 1].token_type == TokenType.KEYWORD:
            self.is_execute = True
            append_commands(self.commands,
                            f"execute store result score {token.string} {DataPack.var_name}")
            return False

        append_commands(self.commands, variable_operation(
            self.command[key_pos:], self.tokenizer, self.lexer.datapack, self.is_execute))
        return True

    def __is_jmc_function(self, key_pos: int, token: Token) -> bool:
        load_once_command = self.get_function(token, LOAD_ONCE_COMMANDS)
        if load_once_command is not None:
            if self.is_execute:
                raise JMCSyntaxException(
                    f"This feature({token.string}) cannot be used with 'execute'", token, self.tokenizer)
            if token.string in self.lexer.datapack.used_command:
                raise JMCSyntaxException(
                    f"This feature({token.string}) can only be used once per datapack", token, self.tokenizer)
            if not self.is_load:
                raise JMCSyntaxException(
                    f"This feature({token.string}) can only be used in load function", token, self.tokenizer)

            self.lexer.datapack.used_command.add(token.string)
            append_commands(self.commands, load_once_command(
                self.command[key_pos + 1], self.lexer.datapack, self.tokenizer).call())
            return True

        execute_excluded_command = self.get_function(
            token, EXECUTE_EXCLUDED_COMMANDS)
        if execute_excluded_command is not None:
            if self.is_execute:
                raise JMCSyntaxException(
                    f"This feature({token.string}) cannot be used with 'execute'", token, self.tokenizer)
            self.lexer.datapack.used_command.add(token.string)
            append_commands(self.commands, execute_excluded_command(
                self.command[key_pos + 1], self.lexer.datapack, self.tokenizer).call())
            return True

        load_only_command = self.get_function(token, LOAD_ONLY_COMMANDS)
        if load_only_command is not None:
            if self.is_execute:
                raise JMCSyntaxException(
                    f"This feature({token.string}) cannot be used with 'execute'", token, self.tokenizer)
            if not self.is_load:
                raise JMCSyntaxException(
                    f"This feature({token.string}) can only be used in load function", token, self.tokenizer)
            append_commands(self.commands, load_only_command(
                self.command[key_pos + 1], self.lexer.datapack, self.tokenizer).call())
            return True

        jmc_command = self.get_function(token, JMC_COMMANDS)
        if jmc_command is not None:
            if len(self.command) > key_pos + 2:
                raise JMCSyntaxException(
                    "Unexpected token", self.command[key_pos + 2], self.tokenizer, display_col_length=False)
            append_commands(self.commands, jmc_command(
                self.command[key_pos + 1], self.lexer.datapack, self.tokenizer, is_execute=self.is_execute).call())
            return True

        if token.string in BOOL_FUNCTIONS:
            raise JMCSyntaxException(
                f"This feature({token.string}) only works in JMC's custom condition", token, self.tokenizer)

        return False

    def __is_flow_control_command(self, key_pos: int, token: Token) -> bool:
        flow_control_command = FLOW_CONTROL_COMMANDS.get(
            token.string, None)
        if flow_control_command is not None:
            if self.is_execute:
                raise JMCSyntaxException(
                    f"This feature({token.string}) cannot be used with 'execute'", token, self.tokenizer)
            return_value = flow_control_command(
                self.command[key_pos:], self.lexer.datapack, self.tokenizer)
            if return_value is not None:
                append_commands(self.commands, return_value)
            return True
        return False

    def get_function(self, token: Token,
                     command_functions: dict[str, type["JMCFunction"]]) -> type["JMCFunction"] | None:
        """
        Get jmc function (class)

        :return: The JMCFunction's subclass
        """
        return command_functions.get(token.string, None)



"""Module containing JMCFunction subclasses for custom JMC function that returns a part of `/execute if` command"""

from ..utils import ArgType
from ..jmc_function import JMCFunction, FuncType, func_property

IF = True
UNLESS = False


@func_property(
    func_type=FuncType.BOOL_FUNCTION,
    call_string='Timer.isOver',
    arg_type={
        "objective": ArgType.KEYWORD,
        "selector": ArgType.SELECTOR
    },
    name='timer_is_over',
    defaults={
        "selector": "@s"
    }
)
class TimerIsOver(JMCFunction):
    def call_bool(self) -> tuple[str, bool]:
        return f'score {self.args["selector"]} {self.args["objective"]} matches 1..', UNLESS



"""Module handling jmc's header"""
from pathlib import Path

from .utils import SingleTon
from .exception import HeaderDuplicatedMacro, HeaderFileNotFoundError, HeaderSyntaxException
from .log import Logger

logger = Logger(__name__)


class Header(SingleTon):
    """
    A SingleTon class containing all information from header
    """
    __slots__ = ('file_read', 'macros', 'credits')

    file_read: set[str]
    """Set of files that was already read (to prevent reading the same file multiple times"""
    macros: dict[str, str]
    """Dictionary of keyword to replace and what to replace it with"""
    credits: list[str]
    """Dictionary of string to replace and what to replace it with"""

    def __init__(self) -> None:
        self.__clear(self)

    @classmethod
    def clear(cls) -> None:
        """
        Reset the single object
        """
        cls.__clear(cls())

    @staticmethod
    def __clear(obj: "Header"):
        obj.file_read = set()
        obj.macros = {}
        obj.credits = []

    def add_file_read(self, path: Path) -> None:
        """
        Add path to file_read

        :param path: Path to file that's read
        """
        self.file_read.add(path.as_posix())

    def is_header_exist(self, path: Path) -> bool:
        """
        Check if header is already in file_read

        :param path: Path to check
        :return: Whether the file was already read
        """
        return path.as_posix() in self.file_read


def parse_header(header_str: str, file_name: str,
                 parent_target: Path) -> Header:
    """
    Parse header and store the information in the header object

    :param header_str: String that was read from the file
    :param file_name: Header file's name
    :param parent_target: Path to parent of the main jmc file
    :raises HeaderSyntaxException: A line in the file doesn't start with '#'
    :raises HeaderDuplicatedMacro: Define same macro twice
    :raises HeaderSyntaxException: Too many/little argument for define
    :raises HeaderSyntaxException: File name isn't wrapped in quote or angle bracket (For `#include`)
    :raises HeaderFileNotFoundError: Can't find header file
    :raises HeaderSyntaxException: Include same file twice
    :raises HeaderSyntaxException: Whitespace found in header file's name
    :raises NotImplementedError: WORKING ON `#replace`
    :raises NotImplementedError: WORKING ON `#credit`
    :raises HeaderSyntaxException: Directive (`#something`) is unrecognized
    :return: Header singleton object

    .. TODO::
        implement replace and credit

    """
    header = Header()
    lines = header_str.split("\n")
    for line, line_str in enumerate(lines):
        line += 1
        if line_str.isspace() or line_str.startswith("//") or line_str == "":
            continue

        if not line_str.startswith("#"):
            raise HeaderSyntaxException(
                "Expected '#'", file_name, line, line_str)

        directive = line_str[1:]
        args = directive.split()

        if args[0] == "define":
            if len(args) == 3:
                key = args[1]
                value = args[2]
                logger.debug(f'Define "{key}" as "{value}"')
                if key in header.macros:
                    raise HeaderDuplicatedMacro(
                        f"'{key}' macro is already defined", file_name, line, line_str)
                header.macros[key] = value
            else:
                raise HeaderSyntaxException(
                    f"'define' takes 2 arguments (got {len(args)-1})", file_name, line, line_str)

        elif args[0] == "include":
            if len(args) == 2:
                included = args[1]
                if (
                    not (included.startswith("<") and included.endswith(">"))
                    and
                    not (included.startswith('"') and included.endswith('"'))
                ):
                    raise HeaderSyntaxException(
                        "Included file must be wrapped in double quotes `\"` or angle bracket `<>`", file_name, line, line_str)
                file_name = included[1:-1]
                if not file_name.endswith(".hjmc"):
                    file_name += ".hjmc"
                header_file = parent_target / file_name
                if not header_file.is_file():
                    raise HeaderFileNotFoundError(header_file)
                with header_file.open('r') as file:
                    header_str = file.read()
                logger.info(f"Parsing {header_file}")
                if header.is_header_exist(header_file):
                    raise HeaderSyntaxException(
                        f"File {header_file.as_posix()} is already included.", file_name, line, line_str)
                header.add_file_read(header_file)
                parse_header(header_str, header_file.as_posix(), parent_target)
            else:
                raise HeaderSyntaxException(
                    "Whitespace is not supported in header file name", file_name, line, line_str)

        elif args[0] == "credit":
            if len(args) == 1:
                header.credits.append("")
            else:
                header.credits.append(" ".join(args[1:]))

        else:
            raise HeaderSyntaxException(
                f"Unrecognized directive '{args[0]}'", file_name, line, line_str)
    return header



"""Module handling logging"""
from io import StringIO
import logging

debug_log_stream = StringIO()
info_log_stream = StringIO()
FORMATTER = logging.Formatter(
    '%(asctime)s | %(name)s | %(levelname)s | %(message)s')


def get_info_log() -> str:
    """Return log string"""
    return info_log_stream.getvalue()


def get_debug_log() -> str:
    """Return log string"""
    return debug_log_stream.getvalue()


def Logger(name: str) -> logging.Logger:
    """Usage:
    ```python
    logger = Logger(__name__)
    ```"""

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    stream_handler = logging.StreamHandler(debug_log_stream)
    stream_handler.setFormatter(FORMATTER)
    stream_handler.setLevel(logging.DEBUG)
    logger.addHandler(stream_handler)

    stream_handler = logging.StreamHandler(info_log_stream)
    stream_handler.setFormatter(FORMATTER)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)

    return logger
